"""Symbol unit tests."""

import math
from finesse import constants
from finesse.symbols import FUNCTIONS, CONSTANTS, Constant


def test_symbols_with_numbers():
    """Test that symbol expressions with numbers evaluate to the correct values."""
    assert (Constant(3.141) + Constant(3.141)).eval() == 6.282
    assert (Constant(3.141) - Constant(3.141)).eval() == 0
    assert (Constant(3.141) * Constant(10)).eval() == 31.41
    assert (Constant(3.141) / Constant(10)).eval() == 0.3141


def test_symbols_with_constants():
    """Test that symbol expressions with constants evaluate to the correct values."""
    assert CONSTANTS["pi"].eval() == constants.PI
    assert CONSTANTS["c0"].eval() == constants.C_LIGHT


def test_symbols_with_functions__single():
    """Test that symbol expressions with single parameter functions evaluate to the
    correct values."""
    assert FUNCTIONS["cos"](0).eval() == 1
    assert FUNCTIONS["cos"](CONSTANTS["pi"]).eval() == -1

    # abs_tol required because math.isclose(anything, 0) is always False by default.
    assert math.isclose(FUNCTIONS["cos"](CONSTANTS["pi"] / 2).eval(), 0, abs_tol=1e-15)
    assert math.isclose(
        FUNCTIONS["cos"](3 * CONSTANTS["pi"] / 2).eval(), 0, abs_tol=1e-15
    )
    assert math.isclose(FUNCTIONS["exp"](1j * CONSTANTS["pi"]).eval().real, -1)


def test_symbols_with_functions__multiple():
    """Test that symbol expressions with multiple parameter functions evaluate to the
    correct values."""
    assert FUNCTIONS["arctan2"](Constant(0), Constant(0)).eval() == 0
    assert FUNCTIONS["arctan2"](Constant(1), Constant(1)).eval() == constants.PI / 4
    assert (
        FUNCTIONS["arctan2"](CONSTANTS["pi"], CONSTANTS["pi"]).eval()
        == constants.PI / 4
    )
    assert (
        FUNCTIONS["arctan2"](Constant(1.23e-5), Constant(6.43e-7)).eval()
        == 1.518567446860085
    )


def test_symbols_with_nested_functions():
    """Test that symbol expressions with nested functions evaluate to the correct
    values."""
    assert FUNCTIONS["arccos"](FUNCTIONS["cos"](CONSTANTS["pi"])).eval() == constants.PI
    assert (
        FUNCTIONS["arcsin"](FUNCTIONS["sin"](CONSTANTS["pi"] / 2)).eval()
        == constants.PI / 2
    )
    # sin²(theta) = (1 - cos(2 theta)) / 2
    assert math.isclose(
        (Constant(1) - FUNCTIONS["cos"](Constant(2) * Constant(8.8)))
        / Constant(2).eval(),
        math.sin(8.8) ** 2,
    )



import numpy as np

from finesse.components import Cavity
from finesse.detectors.general import Detector
from finesse.detectors.compute.gaussian import (
    CPDetectorWorkspace,
    CPDetectorABCDWorkspace,
    CPDetectorModeWorkspace,
    CavityProperty,
)


# Map of cavity property keywords to enum fields.
CP_KEYWORDS = {
    "length": CavityProperty.LENGTH,
    "l": CavityProperty.LENGTH,
    "loss": CavityProperty.LOSS,
    "finesse": CavityProperty.FINESSE,
    "fsr": CavityProperty.FSR,
    "fwhm": CavityProperty.FWHM,
    "pole": CavityProperty.POLE,
    "tau": CavityProperty.TAU,
    "abcd": CavityProperty.ABCD,
    "g": CavityProperty.STABILITY,
    "stability": CavityProperty.STABILITY,
    "gouy": CavityProperty.RTGOUY,
    "modesep": CavityProperty.MODESEP,
    "resolution": CavityProperty.RESOLUTION,
    "q": CavityProperty.EIGENMODE,
    "w": CavityProperty.SOURCE_SIZE,
    "w0": CavityProperty.SOURCE_WAISTSIZE,
    "z": CavityProperty.SOURCE_DISTANCE,
    "zr": CavityProperty.SOURCE_RAYLEIGH,
    "div": CavityProperty.SOURCE_DIVERGENCE,
    "rc": CavityProperty.SOURCE_ROC,
    "s": CavityProperty.SOURCE_DEFOCUS,
}


class CavityPropertyDetector(Detector):
    """Probe for detecting the properties of a cavity.

    The valid values for `prop` are:

      * ``"length"`` or ``"l"``: round-trip cavity length [metres],
      * ``"loss"``: round-trip loss as a fraction,
      * ``"finesse"``: the cavity finesse,
      * ``"fsr"``: free spectral range [Hz],
      * ``"fwhm"``: full-width at half-maximum (i.e. linewidth) [Hz],
      * ``"pole"``: cavity pole frequency [Hz],
      * ``"tau"``: photon storage time [s],
      * ``"abcd"``: round-trip ABCD matrix,
      * ``"g"`` or ``"stability"``: stability as g-factor,
      * ``"gouy"``: round-trip Gouy phase [deg],
      * ``"modesep"``: mode-separation frequency [Hz],
      * ``"resolution"``: cavity resolution [Hz],
      * ``"q"``: eigenmode,
      * ``"w"``: beam size at the cavity source node [metres],
      * ``"w0"``: waist size [metres],
      * ``"z"``: distance to the waist from the cavity source node [metres],
      * ``"zr"``: the Rayleigh range of the eigenmode [metres],
      * ``"div"``: divergence angle of cavity mode [radians],
      * ``"rc"``: radius of curvature of wavefront at cavity source node [metres],
      * ``"s"``: curvature of wavefront at cavity source node [1 / metres].

    Parameters
    ----------
    name : str
        Name of newly created cavity property detector.

    cavity : str or :class:`.Cavity`
        The cavity to probe. If the name is provided then the
        :attr:`.CavityPropertyDetector.cavity` attribute will point
        to the corresponding :class:`.Cavity` object when adding
        this detector to a :class:`.Model` instance.

    prop : str or :class:`.CavityProperty`
        Property of the cavity to probe. See above for options.

    direction : str, optional; default: 'x'
        Plane to detect in.

    q_as_bp : bool, optional; default: False
        If detecting q, should the detector output return :class:`.BeamParam`
        object instead of just a complex number.
    """

    def __init__(self, name, cavity, prop, direction="x", q_as_bp=False):
        if isinstance(prop, str):
            if prop.casefold() not in CP_KEYWORDS:
                raise ValueError(
                    f"Unrecognised property: {prop}, expected "
                    f"one of: {list(CP_KEYWORDS.keys())}"
                )
            prop = CP_KEYWORDS[prop.casefold()]

        units = ""
        if prop == CavityProperty.EIGENMODE:
            if q_as_bp:
                dtype = object
            else:
                dtype = np.complex128
        else:
            dtype = np.float64
            if prop == CavityProperty.RTGOUY:
                units = "degrees"
            elif prop == CavityProperty.SOURCE_DIVERGENCE:
                units = "radians"
            elif prop == CavityProperty.SOURCE_DEFOCUS:
                units = "1/m"
            elif prop == CavityProperty.TAU:
                units = "s"
            elif prop in (
                CavityProperty.FSR,
                CavityProperty.FWHM,
                CavityProperty.POLE,
                CavityProperty.MODESEP,
            ):
                units = "Hz"
            elif prop in (
                CavityProperty.LENGTH,
                CavityProperty.SOURCE_SIZE,
                CavityProperty.SOURCE_WAISTSIZE,
                CavityProperty.SOURCE_DISTANCE,
                CavityProperty.SOURCE_RAYLEIGH,
                CavityProperty.SOURCE_ROC,
            ):
                units = "m"

        if prop == CavityProperty.ABCD:
            shape = (2, 2)
        else:
            shape = None

        property_to_label = {
            CavityProperty.LENGTH: "Cavity round-trip length",
            CavityProperty.LOSS: "Cavity loss",
            CavityProperty.FINESSE: "Cavity finesse",
            CavityProperty.FSR: "Cavity FSR",
            CavityProperty.FWHM: "Cavity FWHM",
            CavityProperty.POLE: "Cavity pole frequency",
            CavityProperty.TAU: "Cavity storage time",
            CavityProperty.ABCD: "Round-trip ABCD matrix",
            CavityProperty.STABILITY: "Stability of cavity",
            CavityProperty.RTGOUY: "Cavity round-trip Gouy phase",
            CavityProperty.MODESEP: "Cavity mode separation frequency",
            CavityProperty.RESOLUTION: "Cavity resolution",
            CavityProperty.EIGENMODE: "Cavity eigenmode",
            CavityProperty.SOURCE_SIZE: "Beam radius of eigenmode",
            CavityProperty.SOURCE_WAISTSIZE: "Waist radius of eigenmode",
            CavityProperty.SOURCE_DISTANCE: "Distance to waist of eigenmode",
            CavityProperty.SOURCE_RAYLEIGH: "Rayleigh range of eigenmode",
            CavityProperty.SOURCE_DIVERGENCE: "Divergence angle of eigenmode",
            CavityProperty.SOURCE_ROC: "Wavefront RoC of eigenmode",
            CavityProperty.SOURCE_DEFOCUS: "Wavefront defocus of eigenmode",
        }
        label = property_to_label[prop]

        Detector.__init__(self, name, dtype=dtype, shape=shape, unit=units, label=label)
        self.__cavity = cavity
        self.__prop = prop
        self.direction = direction

        self.q_as_bp = q_as_bp

    @property
    def prop(self):
        return self.__prop

    @property
    def needs_fields(self):
        return False

    @property
    def needs_trace(self):
        return self.detecting in (
            CavityProperty.EIGENMODE,
            CavityProperty.SOURCE_SIZE,
            CavityProperty.SOURCE_WAISTSIZE,
            CavityProperty.SOURCE_DISTANCE,
            CavityProperty.SOURCE_RAYLEIGH,
            CavityProperty.SOURCE_DIVERGENCE,
            CavityProperty.SOURCE_ROC,
            CavityProperty.SOURCE_DEFOCUS,
        )

    @property
    def detecting(self):
        """The property of the cavity which is being detected.

        :getter: Returns the detected property (read-only).
        """
        return self.__prop

    @property
    def cavity(self):
        """The cavity instance being probed."""
        return self.__cavity

    def _set_cavity(self):
        if not self.has_model:
            raise RuntimeError(f"Bug encountered! No model associated with {self.name}")

        if not isinstance(self.cavity, Cavity):
            cavity = self._model.elements.get(self.cavity, None)
            if cavity is None or not isinstance(cavity, Cavity):
                raise ValueError(f"No cavity of name {self.cavity} in model.")

            self.__cavity = cavity

    def _get_workspace(self, sim):
        if self.needs_trace:
            ws = CPDetectorModeWorkspace(self, sim)
            ws.q_as_bp = self.q_as_bp
        else:
            if self.detecting == CavityProperty.ABCD:
                ws = CPDetectorABCDWorkspace(self, sim)
            else:
                ws = CPDetectorWorkspace(self, sim)

        return ws



"""Testing the structure of TraceForests built for beam tracing routines.

These form the core of all beam tracing code so simply testing that these are structured
correctly is enough to check beam tracing itself in most cases.
"""

import logging
import pytest
import finesse.tracing.ctracer as ctracer
from finesse.exceptions import BeamTraceException
from ...util import assert_trace_trees_equal


##### Simple Fabry-Perot cavity tests #####


@pytest.fixture(scope="module")
def fp_cavity_script():
    return """
    l L0 P=1

    s s0 L0.p1 ITM.p1

    m ITM R=0.99 T=0.01 Rc=-10
    s CAV ITM.p2 ETM.p1 L=1
    m ETM R=0.99 T=0.01 Rc=10
    """


def test_fp_single_cavity_forest(model, fp_cavity_script):
    """Check forest structure for simple FP cavity file with single Cavity object as
    dependency."""
    model.parse(fp_cavity_script)
    model.parse("cav FP ITM.p2")

    model.beam_trace()

    internal_cav_tree = ctracer.TraceTree.from_cavity(model.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        model.ETM.p2.o, model.FP, symmetric=True, pre_node=model.ETM.p1.i,
    )
    external_tree_back = ctracer.TraceTree.from_node(
        model.ITM.p1.o, model.FP, symmetric=True, pre_node=model.ITM.p2.i,
    )

    forest = [internal_cav_tree, external_tree_out, external_tree_back]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_fp_single_gauss_at_laser_forest(model, fp_cavity_script):
    """Check forest structure for simple FP cavity file with single Gauss object,
    positioned at Laser output, as dependency."""
    model.parse(fp_cavity_script)
    model.parse("gauss gL0 L0.p1.o w0=1.2m z=-1.2")

    model.beam_trace()

    gauss_tree = ctracer.TraceTree.from_node(
        model.L0.p1.o, model.gL0, symmetric=True, is_gauss=True,
    )

    forest = [gauss_tree]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_fp_single_gauss_at_laser_reversed_forest(model, fp_cavity_script):
    """Check forest structure for simple FP cavity file with single Gauss object,
    positioned at Laser output but reversed into the Laser, as dependency."""
    model.parse(fp_cavity_script)
    model.parse("gauss gL0 L0.p1.i w0=1.2m z=1.2")

    model.beam_trace()

    gauss_tree_single = ctracer.TraceTree.from_node(
        model.L0.p1.i, model.gL0, symmetric=True, is_gauss=True,
    )
    gauss_tree_forward = ctracer.TraceTree.from_node(
        model.L0.p1.o, model.gL0, symmetric=True, is_gauss=True,
    )

    forest = [gauss_tree_single, gauss_tree_forward]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_fp_cavity_with_laser_gauss_forest__implicit_order(model, fp_cavity_script):
    """Check forest structure for simple FP cavity file with Gauss object, positioned at
    Laser output, and Cavity object as dependencies.

    No tracing priorities given, using implicit ordering.
    """
    model.parse(fp_cavity_script)
    model.parse(
        """
    cav FP ITM.p2
    gauss gL0 L0.p1.o w0=1.2m z=-1.2
    """
    )

    model.beam_trace()

    internal_cav_tree = ctracer.TraceTree.from_cavity(model.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        model.ETM.p2.o, model.FP, symmetric=True, pre_node=model.ETM.p1.i,
    )
    gauss_tree = ctracer.TraceTree.from_node(
        model.L0.p1.o,
        model.gL0,
        symmetric=True,
        is_gauss=True,
        exclude=model.FP.path.nodes_only,
    )

    forest = [internal_cav_tree, external_tree_out, gauss_tree]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_fp_cavity_with_laser_gauss_forest__explicit_order(model, fp_cavity_script):
    """Check forest structure for simple FP cavity file with Gauss object, positioned at
    Laser output, and Cavity object as dependencies.

    Set Gauss object as highest priority.
    """
    model.parse(fp_cavity_script)
    model.parse(
        """
    cav FP ITM.p2
    gauss gL0 L0.p1.o w0=1.2m z=-1.2 priority=1
    """
    )

    model.beam_trace()

    internal_cav_tree = ctracer.TraceTree.from_cavity(model.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        model.ETM.p2.o, model.FP, symmetric=True, pre_node=model.ETM.p1.i,
    )
    gauss_tree = ctracer.TraceTree.from_node(
        model.L0.p1.o,
        model.gL0,
        symmetric=True,
        is_gauss=True,
        exclude=model.FP.path.nodes_only,
    )

    forest = [internal_cav_tree, gauss_tree, external_tree_out]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_fp_single_gauss_at_back_of_cavity(model, fp_cavity_script):
    """Check forest structure for simple FP cavity file with single Gauss object,
    positioned at the rear of the cavity, as dependency."""
    model.parse(fp_cavity_script)
    model.parse("gauss gL0 ETM.p2.i w0=1.2m z=-1.2")

    model.beam_trace()

    gauss_tree = ctracer.TraceTree.from_node(
        model.ETM.p2.i, model.gL0, symmetric=True, is_gauss=True,
    )

    forest = [gauss_tree]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_fp_cavity_with_gauss_at_back(model, fp_cavity_script):
    """Check forest structure for simple FP cavity file with Gauss object, positioned at
    rear of cavity, and Cavity object as dependencies.

    No tracing priorities given, using implicit ordering.
    """
    model.parse(fp_cavity_script)
    model.parse(
        """
    cav FP ITM.p2
    gauss gL0 ETM.p2.i w0=1.2m z=-1.2
    """
    )

    model.beam_trace()

    internal_cav_tree = ctracer.TraceTree.from_cavity(model.FP)
    external_tree_back = ctracer.TraceTree.from_node(
        model.ITM.p1.o, model.FP, symmetric=True, pre_node=model.ITM.p2.i,
    )
    gauss_tree = ctracer.TraceTree.from_node(
        model.ETM.p2.i,
        model.gL0,
        symmetric=True,
        is_gauss=True,
        exclude=model.FP.path.nodes_only,
    )

    forest = [internal_cav_tree, external_tree_back, gauss_tree]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_fp_cavity_with_gauss_in_cavity_logs_error(model, fp_cavity_script, caplog):
    """Test that putting a Gauss object inside a Cavity path logs an error and continues
    without Gauss."""
    model.parse(fp_cavity_script)
    model.parse(
        """
    cav FP ITM.p2
    gauss gL0 ITM.p2.o w0=1.2m z=-1.2
    """
    )

    with caplog.at_level(logging.ERROR):
        model.beam_trace()

        assert "Ignoring this Gauss." in caplog.text

    internal_cav_tree = ctracer.TraceTree.from_cavity(model.FP)
    external_tree_out = ctracer.TraceTree.from_node(
        model.ETM.p2.o, model.FP, symmetric=True, pre_node=model.ETM.p1.i,
    )
    external_tree_back = ctracer.TraceTree.from_node(
        model.ITM.p1.o, model.FP, symmetric=True, pre_node=model.ITM.p2.i,
    )

    forest = [internal_cav_tree, external_tree_out, external_tree_back]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_fp_single_cavity_unstable_raises_exception(model, fp_cavity_script):
    """Test that Fabry-Perot cavity file with cavity unstable raises
    BeamTraceException."""
    model.parse(fp_cavity_script)
    model.elements["CAV"].L = 20
    model.parse("cav FP ITM.p2")

    with pytest.raises(BeamTraceException):
        model.beam_trace()


def test_fp_gauss_at_laser_and_unstable_cavity_forest(model, fp_cavity_script, caplog):
    """Check forest structure for simple FP cavity file with Gauss object, positioned at
    Laser output, as dependency and an unstable cavity."""
    model.parse(fp_cavity_script)
    model.elements["CAV"].L = 20
    model.parse(
        """
    cav FP ITM.p2
    gauss gL0 L0.p1.o w0=1.2m z=-1.2
    """
    )

    with caplog.at_level(logging.WARNING):
        model.beam_trace()

        assert "The cavities ['FP'] are unstable" in caplog.text

    gauss_tree = ctracer.TraceTree.from_node(
        model.L0.p1.o, model.gL0, symmetric=True, is_gauss=True,
    )

    forest = [gauss_tree]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


##### Michelson (no recycling) tests #####


@pytest.fixture(scope="module")
def michelson_fp_script():
    return """
    l L0 P=1
    s s0 L0.p1 BS.p1 L=10

    bs BS

    s sY BS.p2 ITMY.p1 L=100

    m ITMY R=0.99 T=0.01 Rc=-5580
    s LY ITMY.p2 ETMY.p1 L=10k
    m ETMY R=0.99 T=0.01 Rc=5580

    s sX BS.p3 ITMX.p1 L=50

    m ITMX R=0.99 T=0.01 Rc=-5580
    s LX ITMX.p2 ETMX.p1 L=10k
    m ETMX R=0.99 T=0.01 Rc=5580
    """


def test_mfp_forest_xy_arm_cavities__implicit_order(model, michelson_fp_script):
    """Check forest structure for Michelson with both ARM cavities.

    No priorities given so using implicit order.
    """
    model.parse(michelson_fp_script)
    model.parse(
        """
    cav FPX ITMX.p2
    cav FPY ITMY.p2
    """
    )
    model.beam_trace()

    internal_Xcav_tree = ctracer.TraceTree.from_cavity(model.FPX)
    internal_Ycav_tree = ctracer.TraceTree.from_cavity(model.FPY)

    external_Xcav_out = ctracer.TraceTree.from_node(
        model.ETMX.p2.o, model.FPX, symmetric=True, pre_node=model.ETMX.p1.i,
    )
    external_Xcav_back = ctracer.TraceTree.from_node(
        model.ITMX.p1.o, model.FPX, symmetric=True, pre_node=model.ITMX.p2.i,
    )

    external_Ycav_out = ctracer.TraceTree.from_node(
        model.ETMY.p2.o, model.FPY, symmetric=True, pre_node=model.ETMY.p1.i,
    )
    external_Ycav_back = ctracer.TraceTree.from_node(
        model.ITMY.p1.o,
        model.FPY,
        symmetric=True,
        pre_node=model.ITMY.p2.i,
        exclude=[model.BS.p1.o, model.BS.p4.o],
    )

    forest = [
        internal_Xcav_tree,
        internal_Ycav_tree,
        external_Xcav_out,
        external_Xcav_back,
        external_Ycav_out,
        external_Ycav_back,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_mfp_forest_xy_arm_cavities__explicit_order(model, michelson_fp_script):
    """Check forest structure for Michelson with both ARM cavities.

    Set YARM as highest priority.
    """
    model.parse(michelson_fp_script)
    model.parse(
        """
    cav FPX ITMX.p2
    cav FPY ITMY.p2 priority=2
    """
    )
    model.beam_trace()

    internal_Xcav_tree = ctracer.TraceTree.from_cavity(model.FPX)
    internal_Ycav_tree = ctracer.TraceTree.from_cavity(model.FPY)

    external_Xcav_out = ctracer.TraceTree.from_node(
        model.ETMX.p2.o, model.FPX, symmetric=True, pre_node=model.ETMX.p1.i,
    )
    external_Xcav_back = ctracer.TraceTree.from_node(
        model.ITMX.p1.o,
        model.FPX,
        symmetric=True,
        pre_node=model.ITMX.p2.i,
        exclude=[model.BS.p1.o, model.BS.p4.o],
    )

    external_Ycav_out = ctracer.TraceTree.from_node(
        model.ETMY.p2.o, model.FPY, symmetric=True, pre_node=model.ETMY.p1.i,
    )
    external_Ycav_back = ctracer.TraceTree.from_node(
        model.ITMY.p1.o, model.FPY, symmetric=True, pre_node=model.ITMY.p2.i,
    )

    forest = [
        internal_Ycav_tree,
        internal_Xcav_tree,
        external_Ycav_out,
        external_Ycav_back,
        external_Xcav_out,
        external_Xcav_back,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_mfp_forest_gauss_laser(model, michelson_fp_script):
    """Check forest structure for Michelson with neither ARM cavity, but single gauss at
    laser output."""
    model.parse(michelson_fp_script)
    model.parse("gauss gL0 L0.p1.o w0=1.2m z=-1.2")
    model.beam_trace()

    gauss_tree = ctracer.TraceTree.from_node(
        model.L0.p1.o, model.gL0, symmetric=True, is_gauss=True,
    )
    gauss_bs_branch_tree = ctracer.TraceTree.from_node(
        model.BS.p3.i, model.gL0, symmetric=True, exclude=[model.BS.p1.o,],
    )

    forest = [
        gauss_tree,
        gauss_bs_branch_tree,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_mfp_forest_xy_arm_cavities_gauss_laser__implicit_order(
    model, michelson_fp_script
):
    """Check forest structure for Michelson with both ARM cavities and a gauss at the
    laser.

    No priorities given so using implicit order.
    """
    model.parse(michelson_fp_script)
    model.parse(
        """
    cav FPX ITMX.p2.o
    cav FPY ITMY.p2.o
    gauss gL0 L0.p1.o w0=1.2m z=-1.2
    """
    )
    model.beam_trace()

    internal_Xcav_tree = ctracer.TraceTree.from_cavity(model.FPX)
    internal_Ycav_tree = ctracer.TraceTree.from_cavity(model.FPY)

    external_Xcav_out = ctracer.TraceTree.from_node(
        model.ETMX.p2.o, model.FPX, symmetric=True, pre_node=model.ETMX.p1.i,
    )
    external_Xcav_back = ctracer.TraceTree.from_node(
        model.ITMX.p1.o,
        model.FPX,
        symmetric=True,
        pre_node=model.ITMX.p2.i,
        exclude=[model.BS.p1.o],
    )

    external_Ycav_out = ctracer.TraceTree.from_node(
        model.ETMY.p2.o, model.FPY, symmetric=True, pre_node=model.ETMY.p1.i,
    )
    external_Ycav_back = ctracer.TraceTree.from_node(
        model.ITMY.p1.o,
        model.FPY,
        symmetric=True,
        pre_node=model.ITMY.p2.i,
        exclude=[model.BS.p1.o, model.BS.p4.o],
    )
    gtree_exclude = model.FPX.path.nodes_only + model.FPY.path.nodes_only
    gtree_exclude += [model.BS.p2.o, model.ITMY.p1.i, model.BS.p3.o, model.ITMX.p1.i]
    gauss_tree = ctracer.TraceTree.from_node(
        model.L0.p1.o, model.gL0, symmetric=True, is_gauss=True, exclude=gtree_exclude,
    )

    forest = [
        internal_Xcav_tree,
        internal_Ycav_tree,
        external_Xcav_out,
        external_Xcav_back,
        external_Ycav_out,
        external_Ycav_back,
        gauss_tree,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_mfp_forest_xy_arm_cavities_gauss_laser__explicit_order1(
    model, michelson_fp_script
):
    """Check forest structure for Michelson with both ARM cavities and a gauss at the
    laser.

    Gauss has higher priority than both cavities.
    """
    model.parse(michelson_fp_script)
    model.parse(
        """
    cav FPX ITMX.p2.o
    cav FPY ITMY.p2.o
    gauss gL0 L0.p1.o w0=1.2m z=-1.2 priority=1
    """
    )
    model.beam_trace()

    internal_Xcav_tree = ctracer.TraceTree.from_cavity(model.FPX)
    internal_Ycav_tree = ctracer.TraceTree.from_cavity(model.FPY)

    external_Xcav_out = ctracer.TraceTree.from_node(
        model.ETMX.p2.o, model.FPX, symmetric=True, pre_node=model.ETMX.p1.i,
    )

    external_Ycav_out = ctracer.TraceTree.from_node(
        model.ETMY.p2.o, model.FPY, symmetric=True, pre_node=model.ETMY.p1.i,
    )

    gtree_exclude = model.FPX.path.nodes_only + model.FPY.path.nodes_only
    gauss_tree = ctracer.TraceTree.from_node(
        model.L0.p1.o, model.gL0, symmetric=True, is_gauss=True, exclude=gtree_exclude,
    )
    gauss_bs_branch_tree = ctracer.TraceTree.from_node(
        model.BS.p3.i, model.gL0, symmetric=True, exclude=[model.BS.p1.o,],
    )

    forest = [
        internal_Xcav_tree,
        internal_Ycav_tree,
        gauss_tree,
        external_Xcav_out,
        external_Ycav_out,
        gauss_bs_branch_tree,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_mfp_forest_xy_arm_cavities_gauss_laser__explicit_order2(
    model, michelson_fp_script
):
    """Check forest structure for Michelson with both ARM cavities and a gauss at the
    laser.

    XARM cavity has highest priority, followed by Gauss then YARM last as implicit.
    """
    model.parse(michelson_fp_script)
    model.parse(
        """
    cav FPX ITMX.p2.o priority=2
    cav FPY ITMY.p2.o
    gauss gL0 L0.p1.o w0=1.2m z=-1.2 priority=1
    """
    )
    model.beam_trace()

    internal_Xcav_tree = ctracer.TraceTree.from_cavity(model.FPX)
    internal_Ycav_tree = ctracer.TraceTree.from_cavity(model.FPY)

    external_Xcav_out = ctracer.TraceTree.from_node(
        model.ETMX.p2.o, model.FPX, symmetric=True, pre_node=model.ETMX.p1.i,
    )
    external_Xcav_back = ctracer.TraceTree.from_node(
        model.ITMX.p1.o,
        model.FPX,
        symmetric=True,
        pre_node=model.ITMX.p2.i,
        exclude=[model.BS.p1.o],
    )

    external_Ycav_out = ctracer.TraceTree.from_node(
        model.ETMY.p2.o, model.FPY, symmetric=True, pre_node=model.ETMY.p1.i,
    )

    gtree_exclude = model.FPX.path.nodes_only + model.FPY.path.nodes_only
    gtree_exclude += [
        model.BS.p3.o,
    ]
    gauss_tree = ctracer.TraceTree.from_node(
        model.L0.p1.o, model.gL0, symmetric=True, is_gauss=True, exclude=gtree_exclude,
    )

    forest = [
        internal_Xcav_tree,
        internal_Ycav_tree,
        external_Xcav_out,
        external_Xcav_back,
        gauss_tree,
        external_Ycav_out,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_mfp_forest_only_xarm(model, michelson_fp_script):
    """Check forest structure for Michelson with only XARM cavity.

    This tests that branching at beam splitters --- from BS.p1 -> yarm branch.
    """
    model.parse(michelson_fp_script)
    model.parse(
        """
    cav FPX ITMX.p2
    """
    )
    model.beam_trace()

    internal_Xcav_tree = ctracer.TraceTree.from_cavity(model.FPX)

    external_Xcav_out = ctracer.TraceTree.from_node(
        model.ETMX.p2.o, model.FPX, symmetric=True, pre_node=model.ETMX.p1.i,
    )
    external_Xcav_back = ctracer.TraceTree.from_node(
        model.ITMX.p1.o, model.FPX, symmetric=True, pre_node=model.ITMX.p2.i,
    )
    branch_to_yarm_tree = ctracer.TraceTree.from_node(
        model.BS.p1.i, model.FPX, symmetric=True, exclude=[model.BS.p3.o],
    )

    forest = [
        internal_Xcav_tree,
        external_Xcav_out,
        external_Xcav_back,
        branch_to_yarm_tree,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_mfp_forest_only_yarm(model, michelson_fp_script):
    """Check forest structure for Michelson with only YARM cavity.

    This tests that branching at beam splitters --- from BS.p4 -> xarm branch.
    """
    model.parse(michelson_fp_script)
    model.parse(
        """
    cav FPY ITMY.p2
    """
    )
    model.beam_trace()

    internal_Ycav_tree = ctracer.TraceTree.from_cavity(model.FPY)

    external_Ycav_out = ctracer.TraceTree.from_node(
        model.ETMY.p2.o, model.FPY, symmetric=True, pre_node=model.ETMY.p1.i,
    )
    external_Ycav_back = ctracer.TraceTree.from_node(
        model.ITMY.p1.o, model.FPY, symmetric=True, pre_node=model.ITMY.p2.i,
    )
    branch_to_xarm_tree = ctracer.TraceTree.from_node(
        model.BS.p4.i, model.FPY, symmetric=True, exclude=[model.BS.p2.o],
    )

    forest = [
        internal_Ycav_tree,
        external_Ycav_out,
        external_Ycav_back,
        branch_to_xarm_tree,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


##### Linear coupled cavity (with no overlapping) tests #####


@pytest.fixture(scope="module")
def linear_coupled_cav_script():
    return """
    l L0 P=1

    s s0 L0.p1 ITM.p1

    m ITM R=0.99 T=0.01 Rc=-2
    s sc1 ITM.p2 CTM.p1 L=1
    m CTM R=0.99 T=0.01 Rc=2
    s sc2 CTM.p2 ETM.p1 L=2
    m ETM R=0.99 T=0.01 Rc=2.2
    """


def test_linear_coupled_cavity_implicit_ordering(model, linear_coupled_cav_script):
    """Test a linear coupled cavity forest structure for two sub-cavities defined with
    implicit trace ordering."""
    model.parse(linear_coupled_cav_script)
    model.parse(
        """
    cav SUB1 ITM.p2
    cav SUB2 CTM.p2
    """
    )
    model.beam_trace()

    internal_sub1_tree = ctracer.TraceTree.from_cavity(model.SUB1)
    internal_sub2_tree = ctracer.TraceTree.from_cavity(model.SUB2)

    external_sub1_back = ctracer.TraceTree.from_node(
        model.ITM.p1.o, model.SUB1, symmetric=True, pre_node=model.ITM.p2.i,
    )
    external_sub2_out = ctracer.TraceTree.from_node(
        model.ETM.p2.o, model.SUB2, symmetric=True, pre_node=model.ETM.p1.i,
    )

    forest = [
        internal_sub1_tree,
        internal_sub2_tree,
        external_sub1_back,
        external_sub2_out,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


def test_linear_coupled_cavity_explicit_ordering(model, linear_coupled_cav_script):
    """Test a linear coupled cavity forest structure for two sub-cavities defined with
    explicit trace ordering."""
    model.parse(linear_coupled_cav_script)
    model.parse(
        """
    cav SUB1 ITM.p2 priority=-1
    cav SUB2 CTM.p2
    """
    )
    model.beam_trace()

    internal_sub1_tree = ctracer.TraceTree.from_cavity(model.SUB1)
    internal_sub2_tree = ctracer.TraceTree.from_cavity(model.SUB2)

    external_sub1_back = ctracer.TraceTree.from_node(
        model.ITM.p1.o, model.SUB1, symmetric=True, pre_node=model.ITM.p2.i,
    )
    external_sub2_out = ctracer.TraceTree.from_node(
        model.ETM.p2.o, model.SUB2, symmetric=True, pre_node=model.ETM.p1.i,
    )

    forest = [
        internal_sub2_tree,
        internal_sub1_tree,
        external_sub2_out,
        external_sub1_back,
    ]

    assert len(forest) == len(model.trace_forest.forest)

    for tree, ttree in zip(forest, model.trace_forest.forest):
        assert_trace_trees_equal(tree, ttree)


# TODO (sjr) Overlapping cavities forest tests should be in a separate file as they
#            can get quite complicated, so need a whole suite of tests just for these

# TODO (sjr) Asymmetric tracing forest tests are a whole different ball game and also
#            need another file as these require another whole suite of tests


# TODO (sjr) More configurations for this file but main things to focus on are:
#
# - Different cavity types: i.e. triangular and bow-tie cavities
# - Configurations with AR + HR surfaces
# - Full Beamsplitter substrate
# - Telescopes / folded cavities
#



import pytest
from finesse.script.exceptions import KatScriptError


def test_fsig_default(model):
    assert model.fsig.f.value is None


@pytest.mark.parametrize("f", (1, 3.141, 10))
def test_fsig(model, f):
    model.parse(f"fsig({f})")
    assert float(model.fsig.f) == f


@pytest.mark.parametrize("f", (0, 0.0, -1, -3.141, -10, "'hi'", 1 + 2j))
def test_invalid_fsig(model, f):
    with pytest.raises(KatScriptError):
        model.parse(f"fsig({f})")



import pytest
from finesse.script.containers import KatToken
from ..util import IMPLICITLINEEND, ENDMARKER


@pytest.mark.parametrize(
    "token,string",
    (
        ("EQUALS", "="),
        ("PLUS", "+"),
        ("MINUS", "-"),
        ("POWER", "**"),
        ("TIMES", "*"),
        ("FLOORDIVIDE", "//"),
        ("DIVIDE", "/"),
        ("COMMA", ","),
        # FIXME: These throw exceptions because of unbalanced delimiters...
        # ("LBRACKET", "["),
        # ("RBRACKET", "]"),
        # ("LPAREN", "("),
        # ("RPAREN", ")"),
    ),
)
def test_literal(tokenizer, token, string):
    assert list(tokenizer.tokenize(string)) == [
        KatToken(1, 1, 1 + len(string), token, string),
        IMPLICITLINEEND(1, 1 + len(string)),
        ENDMARKER(2),
    ]



"""CLI base group tests."""

from finesse.__main__ import cli as finesse_cli
from .util import sanitized_output


def test_banner(cli):
    """Test printing the banner."""
    cli_result = cli.invoke(finesse_cli, ["--banner"])
    assert "Frequency domain INterferomEter Simulation SoftwarE" in sanitized_output(
        cli_result
    )
    assert cli_result.exit_code == 0



"""
Single-frequency amplitude and phase detector.
"""

import numbers
import logging

import numpy as np

from finesse.detectors.compute.amplitude import ADWorkspace
from finesse.detectors.general import MaskedDetector
from finesse.parameter import float_parameter

LOGGER = logging.getLogger(__name__)


@float_parameter("f", "Frequency", units="Hz")
class AmplitudeDetector(MaskedDetector):
    """Represents an amplitude detector which calculates the amplitude and phase of light fields
    at one frequency.

    Parameters
    ----------
    name : str
        Name of newly created detector.

    node : :class:`.Node`
        Node to read output from.

    f : float
        Frequency of light to detect (in Hz).

    n : int, optional; default: None
        Tangential mode index to probe. Defaults to None such that
        all fields of the given frequency are summed.

    m : int, optional; default: None
        Sagittal mode index to probe. Defaults to None such that
        all fields of the given frequency are summed.
    """

    def __init__(self, name, node, f, n=None, m=None):
        MaskedDetector.__init__(
            self, name, node, dtype=np.complex128, label="Amplitude"
        )

        self.f = f

        if (n is None and m is not None) or (m is None and n is not None):
            raise ValueError(
                f"Error in amplitude detector {self.name}:\n"
                f"    Expected both OR neither of n and m to be specified but "
                f"got n = {n} and m = {m}"
            )

        self.__check_mode_number(n, "n")
        self.__check_mode_number(m, "m")

        self.__n = n
        self.__m = m

    def __check_mode_number(self, k, mode_str):
        if k is not None:
            if (
                not isinstance(k, numbers.Number)
                or k < 0
                or (
                    hasattr(k, "is_integer")
                    and not k.is_integer()
                    and not isinstance(k, numbers.Integral)
                )
            ):
                raise ValueError(
                    f"Error in amplitude detector {self.name}:\n"
                    f"    Expected argument {mode_str} to be None or a positive integer "
                    f"but got {mode_str} = {k}"
                )

    @property
    def n(self):
        """The tangential mode index to probe.

        :getter: Returns the tangential mode index being probed (read-only).
        """
        return self.__n

    @property
    def m(self):
        """The sagittal mode index to probe.

        :getter: Returns the sagittal mode index being probed (read-only).
        """
        return self.__m

    def _get_workspace(self, sim):
        return ADWorkspace(self, sim)



import pytest
from finesse.components.general import Connector
from finesse.script.spec import KatSpec
from finesse.script.compiler import KatCompiler


def resolve(value):
    if isinstance(value, list):
        for index, item in enumerate(value):
            value[index] = resolve(item)
    else:
        try:
            value = value.eval()
        except AttributeError:
            pass

    return value


@pytest.fixture(scope="package")
def fuzz_value_parse_compare():
    """Special package-scoped parser for fuzzing values."""

    class FakeNoParamElement(Connector):
        """A fake element with no model parameters."""

        def __init__(self, name, a=None):
            super().__init__(name)
            self.a = a

    # Register special fuzz component.
    spec = KatSpec()
    spec.register_element(("fuzz",), {"setter": FakeNoParamElement})
    compiler = KatCompiler(spec=spec)

    def _(value, expected, exact=True):
        model = compiler.compile(f"fuzz fuzzer {value}")
        value = resolve(model.fuzzer.a)
        expected = resolve(expected)

        if exact:
            assert value == expected
        else:
            assert value == pytest.approx(expected)

    return _



import pytest
from finesse.model import Model
from finesse.components.general import Connector
from finesse.components.node import NodeDirection, NodeType
from finesse.detectors.general import Detector
from finesse.analysis.actions import Action
from finesse.parameter import float_parameter
from finesse.script.spec import BaseSpec


@pytest.fixture
def spec():
    """Kat spec with no registered elements, commands, etc.

    The spec is shared by builder and generator tests.
    """

    class EmptyTestSpec(BaseSpec):
        """Empty specification class, used by tests."""

    return EmptyTestSpec()


@pytest.fixture
def model():
    class FakeModel(Model):
        """A fake model."""

    return FakeModel()


@pytest.fixture
def fake_element_cls():
    @float_parameter("a", "Fake Parameter A")
    @float_parameter("b", "Fake Parameter B")
    class FakeElement(Connector):
        """A fake 2-port element."""

        def __init__(self, name, a=None, b=None):
            super().__init__(name)
            self.a = a
            self.b = b

            # Add some ports.
            self._add_port("p1", NodeType.OPTICAL)
            self.p1._add_node("i", NodeDirection.INPUT)
            self.p1._add_node("o", NodeDirection.OUTPUT)

            self._add_port("p2", NodeType.OPTICAL)
            self.p2._add_node("i", NodeDirection.INPUT)
            self.p2._add_node("o", NodeDirection.OUTPUT)

    return FakeElement


@pytest.fixture
def fake_element_noparam_cls():
    class FakeNoParamElement(Connector):
        """A fake element with no model parameters."""

        def __init__(self, name, a=None, b=None):
            super().__init__(name)
            self.a = a
            self.b = b

            # Add some ports.
            self._add_port("p1", NodeType.OPTICAL)
            self.p1._add_node("i", NodeDirection.INPUT)
            self.p1._add_node("o", NodeDirection.OUTPUT)

            self._add_port("p2", NodeType.OPTICAL)
            self.p2._add_node("i", NodeDirection.INPUT)
            self.p2._add_node("o", NodeDirection.OUTPUT)

    return FakeNoParamElement


@pytest.fixture
def fake_binop_cls():
    class FakeBinaryOperation:
        """A fake binary operation."""

        def __init__(self, operator, lhs, rhs):
            self.operator = operator
            self.lhs = lhs
            self.rhs = rhs

        def eval(self):
            return eval(f"{self.lhs}{self.operator}{self.rhs}")

    return FakeBinaryOperation


@pytest.fixture
def fake_unop_cls():
    class FakeUnaryOperation:
        """A fake unary operation."""

        def __init__(self, operator, argument):
            self.operator = operator
            self.argument = argument

        def eval(self):
            return eval(f"{self.operator}{self.argument}")

    return FakeUnaryOperation


@pytest.fixture
def fake_detector_cls():
    class FakeDetector(Detector):
        """A fake detector element."""

        def __init__(self, name, target):
            super().__init__(name)
            self.target = target

    return FakeDetector


@pytest.fixture
def fake_analysis_cls():
    class FakeAnalysis(Action):
        """A fake action."""

        def __init__(self, a=None, b=None):
            self.a = a
            self.b = b

        def _do(self, state):
            raise NotImplementedError

        def _requests(self, model, memo, first=True):
            raise NotImplementedError

    return FakeAnalysis



"""CLI info command tests."""

from finesse.__main__ import info
from .util import sanitized_output


def test_info(cli, input_file):
    """Test info output."""
    cli_result = cli.invoke(info, [input_file])
    assert "Summary:" in sanitized_output(cli_result)
    assert cli_result.exit_code == 0



# pylint: disable=no-member

"""Model path tests."""

import pytest
from networkx.exception import NetworkXNoPath
from finesse.components import Laser, Mirror, Beamsplitter


def test_optical_path_lengths(model):
    """Test path lengths are the sum of the lengths of its spaces."""
    l1 = Laser("lsr")
    m1 = Mirror("m1")
    m2 = Mirror("m2")
    m3 = Mirror("m3")
    model.chain(l1, 0.5, m1, 3.25, m2, 10.5, m3)
    path = model.path(l1.p1.o, m3.p1.i)
    assert path.length == 14.25


def test_optical_path_lengths__non_unity_nr(model):
    """Test path lengths account for index of refraction."""
    l1 = Laser("lsr")
    m1 = Mirror("m1")
    m2 = Mirror("m2")
    m3 = Mirror("m3")
    model.chain(l1, {"L": 3.0, "nr": 1.5}, m1, 3.25, m2, {"L": 10.5, "nr": 2}, m3)
    path = model.path(l1.p1.o, m3.p1.i)
    assert path.length == 10.5


def test_optical_path_via_node_not_in_path_raises_exception(model):
    """Test that specifying a via not on the path causes an exception."""
    # Triangular cavity with each beamsplitter's p1/p2 sides connected.
    m1 = Beamsplitter("m1", T=500e-6, L=0, Rc=12)
    m2 = Beamsplitter("m2", T=15e-6, L=0)
    m3 = Beamsplitter("m3", T=15e-6, L=0)
    model.connect(m1.p1, m2.p2, L=6)
    model.connect(m2.p1, m3.p2, L=6)
    model.connect(m3.p1, m1.p2, L=6)

    # m1
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m1.p1.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m1.p3.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m1.p3.o)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m1.p4.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m1.p4.o)
    # m2
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m2.p1.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m2.p2.o)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m2.p3.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m2.p3.o)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m2.p4.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m2.p4.o)
    # m3
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m3.p1.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m3.p2.o)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m3.p3.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m3.p3.o)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m3.p4.i)
    with pytest.raises(NetworkXNoPath):
        model.path(m1.p1.o, m1.p2.i, via_node=m3.p4.o)



import finesse


def test_angular_spring_phi_dependence():
    # This really just tests the model runs, I don't have an analytic solution
    # for the TF to test against.
    IFO = finesse.Model()

    IFO.parse(
        """
        l l1 P=1k
        s s1 l1.p1 m1.p1 L=0
        m m1 R=0.99 T=0.01 Rc=-2076
        s s2 m1.p2 m2.p1 L=4000
        m m2 R=1 T=0 Rc=2076

        cav c1 m2.p1

        pendulum m1_sus m1.mech I_pitch=1 fpitch=0.6 Qpitch=1000 I_yaw=1 fyaw=0.6 Qyaw=1000
        pendulum m2_sus m2.mech I_pitch=1 fpitch=0.6 Qpitch=1000 I_yaw=1 fyaw=0.6 Qyaw=1000

        pd P m1.p2.o
        ad y1 m1.mech.yaw &fsig.f
        ad y2 m2.mech.yaw &fsig.f
        ad a00 m2.p1.o f=&fsig.f n=0 m=0
        ad a10 m2.p1.o f=&fsig.f n=0 m=1

        sgen s m2.mech.F_yaw

        fsig(1)
        modes(maxtem=1)
        xaxis(s.f, log, 0.01, 10, 1000)
        """
    )
    out1 = IFO.run()
    # Shift cavity but length still on resonance
    # result should be the same
    IFO.m1.phi = 90
    IFO.m2.phi = 90
    out2 = IFO.run()
    # if this isn't true then the signal sidebands are not
    # picking up the correct tuning phase
    assert all(abs(out1["y1"] - out2["y1"]) / abs(out1["y1"]) < 1e-11)
    assert all((abs(out1["y2"] - out2["y2"]) / abs(out1["y2"])) < 1e-11)



"""
The ``components`` module contains all the component type of an interferometer configuration
including the general objects required to connect them and register node connections.

These include not only optical components such as mirrors and lasers but also electrical and
mechanical component types found in physical interferometers.
"""

from .general import Connector, FrequencyGenerator, Variable
from .surface import Surface
from .node import Node, NodeType, NodeDirection, Port
from .beamsplitter import Beamsplitter
from .cavity import Cavity
from .gauss import Gauss
from .directional_beamsplitter import DirectionalBeamsplitter
from .optical_bandpass import OpticalBandpassFilter
from .isolator import Isolator
from .laser import Laser
from .lens import Lens
from .mirror import Mirror
from .modulator import Modulator
from .nothing import Nothing
from .readout import Readout, ReadoutDC, ReadoutRF, ReadoutDCQPD
from .signal import SignalGenerator
from .space import Space
from .squeezer import Squeezer
from .wire import Wire
from .mechanical import Joint
from .electronics import (
    ZPKNodeActuator,
    Amplifier,
    Filter,
    ZPKFilter,
    ButterFilter,
    Cheby1Filter,
)
from .dof import DegreeOfFreedom

__all__ = (
    "Connector",
    "FrequencyGenerator",
    "Variable",
    "Surface",
    "Node",
    "NodeType",
    "NodeDirection",
    "Port",
    "Beamsplitter",
    "Cavity",
    "Gauss",
    "DirectionalBeamsplitter",
    "OpticalBandpassFilter",
    "Isolator",
    "Laser",
    "Lens",
    "Mirror",
    "Modulator",
    "Nothing",
    "Readout",
    "ReadoutDC",
    "ReadoutDCQPD",
    "ReadoutRF",
    "SignalGenerator",
    "Space",
    "Squeezer",
    "Wire",
    "Joint",
    "ZPKNodeActuator",
    "Amplifier",
    "Filter",
    "ZPKFilter",
    "ButterFilter",
    "Cheby1Filter",
    "DegreeOfFreedom",
    "Readout",
)



import pytest
from finesse.script.generator import KatUnbuilder, KatUnfiller
from finesse.script.graph import KatGraph


@pytest.fixture
def unbuilder(spec):
    return KatUnbuilder(spec=spec)


@pytest.fixture
def unfiller():
    return KatUnfiller()


@pytest.fixture
def graph(spec):
    return KatGraph()



"""
Top-level objects which specific detectors should inherit from.
"""

from abc import ABC

import numpy as np

from finesse.element import ModelElement
from finesse.components import Node
from finesse.utilities.homs import make_modes, insert_modes, remove_modes


class Detector(ABC, ModelElement):
    """Abstract representation of a component that produces a numerical output.

    User detector classes should subclass this class. The simulation will then generate a dictionary
    of output values.

    Parameters
    ----------
    name : str
        Name of newly created detector.

    node : :class:`.Node`
        Node to read output from.

    dtype : type, optional
        The numpy datatype for which this output result will be stored in.

    unit : str, optional
        A human readable unit for the output. E.g. W, m, m/rtHz.
    """

    def __init__(
        self,
        name,
        node=None,
        dtype=np.complex128,
        shape=(),
        unit="arb.",
        label=None,
        needs_fields=True,
        needs_trace=False,
    ):
        from finesse.detectors.workspace import OutputInformation

        ModelElement.__init__(self, name)

        if node is not None and not isinstance(node, Node):
            msg = (
                "Expected argument 'node' to be None or of type Node but "
                f"got argument of type {type(node)}"
            )
            raise TypeError(msg)

        if node:
            node.used_in_detector_output.append(self)

        self.__output_information = OutputInformation(
            name, (node,), dtype, unit, shape, label, needs_fields, needs_trace
        )

    def _on_remove(self):
        self.node.used_in_detector_output.remove(self)

    def _set_dtype(self, dtype):
        self.__dtype = dtype

    @property
    def output_information(self):
        return self.__output_information

    @property
    def needs_fields(self):
        """Flag indicating whether the detector requires light fields (i.e. solving
        of the interferometer matrix).
        """
        return self.__output_information.needs_fields

    @property
    def needs_trace(self):
        """Flag indicating whether the detector requires beam traces."""
        return self.__output_information.needs_trace

    @property
    def node(self):
        """The nodes this detector observes.

        :getter: Returns the detected node.
        """
        return self.__output_information.nodes[0]

    @property
    def dtype(self):
        return self.__output_information.dtype

    @property
    def dtype_shape(self):
        return self.__output_information.dtype_shape

    def _update_dtype_shape(self, shape):
        """Only to be used internally by detectors like Cameras
        for updating the shape if resolution is changed."""
        self.__output_information._update_dtype_shape(shape)

    @property
    def dtype_size(self):
        """Size of the output in terms of number of elements.

        This is typically unity as most detectors return a single
        value via their output functions.

        Equivalent to the product of :attr:`.Detector.dtype_shape`.
        """
        return int(np.prod(self.__output_information.dtype_shape))

    @property
    def unit(self):
        return self.__output_information.unit

    @property
    def label(self):
        return self.__output_information.label


class MaskedDetector(Detector, ABC):
    """An abstract class from which detector types which can have mode masks derive.

    Any detector object which calculates quantities involving loops over the
    modes of a model should inherit from this --- allowing masks to be applied
    to mode patterns via the methods of this class. Examples of detectors which
    should derive from `MaskedDetector` are power-detectors, amplitude-detectors
    and cameras.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # The mode indices which will be masked
        # on this detector in simulations
        self._mask = np.zeros(0, dtype=(np.intc, 2))

    @property
    def has_mask(self):
        """Whether the detector has a mask applied to it.

        Simply checks to see if the :attr:`.MaskedDetector.mask`
        length is non-zero.
        """
        return len(self.mask)

    @property
    def mask(self):
        """An array of HOMs to mask from the output. Any contributions from these
        modes will be zeroed when calculating the detector output.

        :getter: Returns the array of masked indices.
        :setter: Sets the masked indices. See :meth:`.MaskedDetector.select_mask` for
                 the options available.
        """
        return self._mask

    @mask.setter
    def mask(self, value):
        self.select_mask(value)

    def select_mask(self, modes=None, maxtem=None, exclude=None):
        """Select the HOM indices to include in the mask.

        The mode-selecting examples in :ref:`selecting_modes` may be referred
        to for typical patterns when using this method, as the same concepts
        apply equally to making detector masks (equivalent code under-the-hood).

        Parameters
        ----------
        modes : sequence, str, optional; default: None
            Identifier for the mode indices to generate. This can be:

            - An iterable of mode indices, where each element in the iterable must unpack to two
              integer convertible values.

            - A string identifying the type of modes to include, must be one of "even", "odd",
              "tangential" (or "x") or "sagittal" (or "y").

            By default this is None, such that, for example, this method can be used to select a
            mask of all modes up to a given `maxtem`.

        maxtem : int, optional; default: None
            Optional maximum mode order. If not specified then the maxtem used internally will be
            equal to the maximum mode order of the associated model.

            Note that this argument is ignored if `modes` is an iterable of mode indices.

        exclude : sequence, str, optional; default: None
            A mode, or iterable of modes, to exclude from the selected pattern. For example, if
            one calls ``select_mask("even", exclude="00")`` then the mask will be an array of
            all even-order HOM indices excluding the 00 mode.

        Examples
        --------
        See :ref:`selecting_modes`.
        """
        if maxtem is None:
            if self.has_model:
                maxtem = self._model.modes_setting["maxtem"]

        self._mask = make_modes(modes, maxtem)

        if exclude is not None:
            self.remove_from_mask(exclude)

    def add_to_mask(self, modes):
        """Inserts the specified mode indices into the detector mask.

        Parameters
        ----------
        modes : sequence, str
            A single mode index pair or an iterable of mode indices. Each
            element must unpack to two integer convertible values.
        """
        self._mask = insert_modes(self._mask, modes)

    def remove_from_mask(self, modes):
        """Removes the specified mode indices from the detector mask.

        Parameters
        ----------
        modes : sequence, str
            A single mode index pair or an iterable of mode indices. Each
            element must unpack to two integer convertible values.
        """
        self._mask = remove_modes(self._mask, modes)


class SymbolDetector(Detector):
    def __init__(self, name, symbol, dtype=np.complex128):
        Detector.__init__(self, name, None, dtype=dtype)
        self.symbol = symbol

    def get_output(self, *args):
        return self.dtype(self.symbol.eval())


class NoiseDetector:
    def __init__(self, noise_type):
        self.__noise_type = noise_type
        self._requested_selection_vectors = dict()
        self.__selection_vectors = {}

    @property
    def noise_type(self):
        return self.__noise_type

    def _request_selection_vector(self, name):
        self._requested_selection_vectors[name] = -1



"""Parameter unit tests."""

from finesse.symbols import Constant


def test_parameter_equality_checks(model):
    """Test that parameters can be directly compared to numbers and other symbols"""
    model.parse("laser l1 P=3.14")
    assert model.l1.P == 3.14
    assert model.l1.P == Constant(3.14)
    assert model.l1.P == (Constant(3.14) + Constant(3.14)) / 2
    assert model.l1.P.ref == model.l1.P


def test_parameter_ref_substitution(model):
    """Test that parameter refs can be subsituted for in expressions"""
    model.parse("laser l1 P=3.14")
    assert model.l1.P.eval() == 3.14
    assert (model.l1.P.ref + 1).eval(subs={model.l1.P: 4}) == 5



# flake8: noqa

"""Finesse is a Python package for simulating interferometers in the frequency
domain."""

import sys
import locale


PROGRAM = __name__
DESCRIPTION = "Simulation program for laser interferometers."

# Set the locale to the user's default (required for e.g. number formatting in log warnings).
locale.setlocale(locale.LC_ALL, "")

# Set the Finesse version.
try:
    from .version import version as __version__
except ImportError:
    raise Exception("Could not find version.py. Ensure you have run setup.")

# Set up some sensible default runtime options.
from .config import configure, autoconfigure

autoconfigure()

# Import a bunch of useful functions and classes into the top-level package.
from .environment import is_interactive, show_tracebacks, tb
from .constants import values as constants
from .parameter import Parameter, float_parameter
from .gaussian import BeamParam
from .model import Model
from .enums import SpatialType
from .plotting import init as init_plotting



"""Kat script to Finesse object adapters.

Adapters provide various useful information to the kat script compiler about Finesse
objects, and vice versa for the kat script generator.
"""

import abc
import inspect
import logging

LOGGER = logging.getLogger(__name__)


def _remove_signature_parameters(signature, remove_first=True, remove_extra=None):
    """Remove any parameter whose name appears in `remove` from `signature`.

    There is also the option to remove the first parameter, which is usually "self" (in
    the case of methods) or some parameter that receives the container (e.g. a
    :class:`.Model` in the case of commands).
    """
    if not remove_extra:
        remove_extra = []
    parameters = []
    for i, (name, parameter) in enumerate(signature.parameters.items()):
        if i == 0 and remove_first:
            continue
        if name in remove_extra:
            continue
        parameters.append(parameter)
    return signature.replace(parameters=parameters)


class BaseAdapter(metaclass=abc.ABCMeta):
    """Adapter defining a kat script instruction and how it maps to/from a type.

    This encapsulates the required information to take a kat script instruction and
    generate a corresponding Python object (e.g. a :class:`.Laser` from a `laser l1 ...`
    instruction) and to dump that Python object back to kat script.

    This is an abstract class that should be subclassed.

    Parameters
    ----------
    aliases : str or sequence
        The instruction alias(es). The first is considered the full name and is dumped
        in full archival mode. The last is considered the short form name and used in
        default dump mode. Any other specified aliases are only supported for parsing.

    setter : type
        The Python setter type for this instruction. This is used to build Python
        objects for elements, analyses, etc., and to set values in the model for
        commands.

    getter : type or callable, optional
        The Python getter type for this instruction. This is used to retrieve the
        object's current parameter values to generate its kat script representation. If
        `getter` is not specified, it is assumed that this instruction should not be
        regenerated in kat script. This may be the same as `setter` or alternatively a
        class that inherts :class:`.GetterProxy`. In the latter case the object should
        implement a `__call__` method that accepts the containing object and returns
        either a single (args, kwargs) tuple or a sequence thereof, depending on
        `singular`.

    singular : bool
        Whether the instruction can only be specified once per script. In such a case,
        if `getter` is a `GetterProxy`, it should only define a single `(args, kwargs)`
        tuple instead of a sequence thereof.

    build_last : bool, optional
        Whether to build the Python object last, regardless of dependencies. This is
        useful for elements with implicit dependencies (see e.g. the cavity adapter). Be
        careful using this flag because statements for other adapters that depend on
        statements for adapters with this flag will be built first. Defaults to False.

    Raises
    ------
    ValueError
        If `aliases` sequence has less than 1 entry.
    """

    def __init__(
        self, aliases, setter, getter=None, singular=False, build_last=False,
    ):
        if isinstance(aliases, str):
            aliases = [aliases]

        if len(aliases) < 1:
            raise ValueError("At least one alias must be specified in 'aliases'")

        self.aliases = aliases
        self.setter = setter
        self.getter = getter
        self.singular = singular
        self.build_last = build_last

    @property
    def full_name(self):
        return self.aliases[0]

    @property
    def short_name(self):
        return self.aliases[-1]

    @abc.abstractmethod
    def apply(self, model, *args, **kwargs):
        raise NotImplementedError

    def get(self, container):
        """Get ordered mapping of argument names to values from `container`.

        Parameters
        ----------
        container : object
            The container to retrieve argument names and values for.

        Returns
        -------
        :class:`list`
            Positional argument values.

        :class:`dict`
            Mapping of keyword argument names to values.

        Raises
        ------
        RuntimeError
            If this adapter's :meth:`.dump_signature` contains parameters that are not
            present in `container`.
        """
        if not self.getter:
            LOGGER.debug(f"{self!r} has no kat script representation")
            return

        if isinstance(self.getter, GetterProxy):
            params = self.getter(container)

            if params is None:
                LOGGER.debug(
                    f"skipping serialisation of empty {self!r} parameters (returned "
                    f"from a GetterProxy)"
                )
                return

            return params
        else:
            kwargs = {}
            for arg, param in self.dump_signature().parameters.items():
                name = param.name
                try:
                    # Try to get the attribute value.
                    value = getattr(container, name)
                except AttributeError as e:
                    raise RuntimeError(
                        f"The dump signature (getter) for '{self!r}' defines parameter "
                        f"'{name}' that is not a property or attribute of "
                        f"'{container!r}'. Either define "
                        f"'{container.__class__.__name__}.{name}' as a property or "
                        f"attribute or set {self.full_name}'s `getter` to a "
                        f"`GetterProxy` object (see BaseAdapter docstring)."
                    ) from e

                kwargs[name] = value

        # FIXME: in the future, this should support being able to return parameters in
        # different modes, like positional or keyword-only.
        return [], kwargs

    @property
    def docobj_type(self):
        """Type to use to get the Python object's docstring.

        :getter: Returns the type to be used to retrieve the Python object's docstring.

        Notes
        -----
        This assumes that classes contain the documentation for the `__init__` method,
        in line with the Numpydoc convention.
        """
        signature_type = self.setter

        if isinstance(signature_type, property):
            # Use the setter's signature.
            signature_type = signature_type.fset

        return signature_type

    def doc_signature(self, exclude_name=False):
        """The signature of the documentation object.

        This is an ordered mapping of the Python object's supported parameters to
        :py:class:`inspect.Parameter` objects.

        Parameters
        ----------
        exclude_name : bool, optional
            Exclude the `name` parameter. This is useful for code that already handles
            the name separately. Defaults to `False`.

        Returns
        -------
        :py:class:`inspect.Signature`
            The documentation object's pseudo-signature.
        """
        return _remove_signature_parameters(
            inspect.signature(self.docobj_type),
            remove_first=True,
            remove_extra=["name"] if exclude_name else [],
        )

    @property
    def call_signature_type(self):
        """Type to use to get the Python object's call signature.

        :getter: Returns the type to be used to retrieve the Python object's call
                 signature.
        """
        signature_type = self.setter

        if inspect.isclass(signature_type):
            # Use the init method.
            signature_type = signature_type.__init__
        elif isinstance(signature_type, property):
            # Use the setter's signature.
            signature_type = signature_type.fset

        return signature_type

    def call_signature(self, exclude_name=False):
        """The call signature of the corresponding Python object constructor.

        This is an ordered mapping of the Python object's supported parameters to
        :py:class:`inspect.Parameter` objects.

        Parameters
        ----------
        exclude_name : bool, optional
            Exclude the `name` parameter. This is useful for code that already handles
            the name separately. Defaults to `False`.

        Returns
        -------
        :py:class:`inspect.Signature`
            The call object's pseudo-signature.
        """
        return _remove_signature_parameters(
            inspect.signature(self.call_signature_type),
            remove_first=True,
            remove_extra=["name"] if exclude_name else [],
        )

    @property
    def dump_signature_type(self):
        signature_type = self.getter

        if isinstance(signature_type, GetterProxy):
            # You're doing it wrong.
            raise TypeError(
                "Adapter has a getter proxy set, so no dump signature type exists. "
                "Change the code calling this method to first check the getter type."
            )

        if inspect.isclass(signature_type):
            # Use the init method.
            signature_type = signature_type.__init__
        elif isinstance(signature_type, property):
            # Use the getter's signature.
            signature_type = signature_type.fget

        return signature_type

    def dump_signature(self, exclude_name=False):
        """The Python object constructor call signature as available to the generator.

        This is used when generating kat script for the corresponding Python object and
        may exclude or include parameters found or not found in the real Python object
        constructor, e.g. to allow a kat script command to use different arguments to
        that of the Python API.

        Parameters
        ----------
        exclude_name : bool, optional
            Exclude the `name` parameter. This is useful for code that already handles
            the name separately. Defaults to `False`.

        Returns
        -------
        :py:class:`inspect.Signature`
            The dump object's pseudo-signature.
        """
        return _remove_signature_parameters(
            inspect.signature(self.dump_signature_type),
            remove_first=True,
            remove_extra=["name"] if exclude_name else [],
        )

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.full_name} @ {hex(id(self))}>"


class ModelObject(BaseAdapter, metaclass=abc.ABCMeta):
    """Mixin for adapters that control Python objects directly (e.g. elements) as
    opposed to attributes (e.g. commands)."""

    def compile(self, args, kwargs):
        return self.setter(*args, **kwargs)


class ElementAdapter(ModelObject, BaseAdapter):
    """Adapter for elements."""

    def apply(self, model, item):
        model.add(item)

    def get(self, item):
        # Split the name argument onto its own.
        args, kwargs = super().get(item)
        element_name = kwargs.pop("name")
        return element_name, args, kwargs


class CommandAdapter(BaseAdapter):
    """Adapter for commands.

    Commands set properties of a :class:`.Model`.

    Command getters should return a *sequence* of args and kwargs representing
    potentially multiple commands to dump.
    """

    def apply(self, model, allargs):
        # The compiler passes this method an (args, kwargs) tuple when building
        # commands.
        args, kwargs = allargs
        return self.setter(model, *args, **kwargs)


class AnalysisAdapter(ModelObject, BaseAdapter):
    """Adapter for analyses."""

    def apply(self, model, item):
        model.analysis = item


class GetterProxy(metaclass=abc.ABCMeta):
    """An object that when called returns the parameters and values that should be
    dumped to represent an object in kat script.

    Inheriting classes should define `__call__`, returning an ordered mapping of
    parameters to values.
    """

    @abc.abstractmethod
    def __call__(self, item):
        raise NotImplementedError



"""Kat file parser number parsing fuzzing tests."""

import numpy as np
from hypothesis import given, assume, settings
from hypothesis.strategies import integers, floats, complex_numbers


_INT_DTYPE_INFO = np.iinfo(np.int32)
MIN_INT = _INT_DTYPE_INFO.min
MAX_INT = _INT_DTYPE_INFO.max


# Need bounds on integers due to Numpy. See #119.
@given(value=integers(min_value=MIN_INT, max_value=MAX_INT))
@settings(deadline=1000)  # See #292.
def test_integer_fuzzing(fuzz_value_parse_compare, value):
    fuzz_value_parse_compare(value, value)


@given(value=floats(allow_nan=False, allow_infinity=True))
@settings(deadline=1000)  # See #292.
def test_float_fuzzing(fuzz_value_parse_compare, value):
    assume(not str(value).endswith("e"))
    fuzz_value_parse_compare(value, value)


@given(value=complex_numbers(allow_nan=False, allow_infinity=False))
@settings(deadline=1000)  # See #292.
def test_complex_fuzzing(fuzz_value_parse_compare, value):
    fuzz_value_parse_compare(value, value)



"""Finesse environment information."""

import sys

try:
    from IPython.core.ultratb import AutoFormattedTB
except ImportError:
    AutoFormattedTB = None

from .datastore import invalidate, init_singleton

# Platform detection.
IS_WINDOWS = sys.platform.startswith("win")


def traceback_handler_instance():
    return init_singleton(_TracebackHandler)


def reset_traceback_handler():
    invalidate(_TracebackHandler)


def is_interactive():
    """Check if Finesse is being run from an interactive environment.

    Returns
    -------
    bool
        `True` if a Jupyter notebook, Qt console or IPython shell is in use, `False`
        otherwise.
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return True  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def show_tracebacks(show_tb=None):
    """Get or set whether to show tracebacks in Jupyter/IPython in full form or just as
    a single error message.

    Parameters
    ----------
    show_tb : bool, optional
        Set whether to show tracebacks; defaults to None, which doesn't change the
        current setting.

    Returns
    -------
    bool
        The current setting value (if `show_tb` was not set) or the previous setting
        value (if `show_tb` was set).
    """
    tb = traceback_handler_instance()
    previous = tb.show_tb
    if show_tb is not None:
        tb.show_tb = show_tb
    return previous


def tb():
    """Show the last traceback in full."""
    tb = traceback_handler_instance()
    tb.show_last()


def has_pygraphviz():
    """Determine if pygraphviz and graphviz are available on the current path.

    Returns
    -------
    bool
        `True` if pygraphviz and graphviz are available, `False` otherwise.

    Notes
    -----

    This returns `False` if either pygraphviz or graphviz is not installed or correctly
    configured. If pygraphviz is installed but graphviz is not installed or could not be
    found, an import error complaining about something like "libagraph.so.1" is emitted
    (see the `pygraphviz FAQ
    <https://pygraphviz.github.io/documentation/stable/reference/faq.html>`_).
    """
    try:
        import pygraphviz  # noqa: F401
    except ImportError:
        return False
    return True


class _TracebackHandler:
    """Handler for user errors during parse and build.

    This detects the environment within which the user is running Finesse - either a
    normal Python terminal or script or within IPython or JupyterLab. Depending on the
    environment, either complete tracebacks are shown or else only the error message.

    The behaviour of this handler can be configured globally using
    :func:`.show_tracebacks`.

    Do not instantiate this class directly; use :func:`traceback_handler_instance`.
    """

    def __init__(self):
        self.show_tb = True
        self.etype = None
        self.evalue = None
        self.tb = None
        self.stb = None
        self.text = None
        self.a = None

        if AutoFormattedTB is not None:
            self.a = AutoFormattedTB(
                mode="Context", color_scheme="Neutral", tb_offset=1
            )

    def store_tb(self):
        import traceback

        self.etype, self.evalue, self.tb = sys.exc_info()

        if AutoFormattedTB is not None:
            self.stb = self.a.structured_traceback(
                self.etype, self.evalue, self.tb, tb_offset=1
            )
            self.text = self.a.stb2text(self.stb)
        else:
            self.text = traceback.format_exc()

    def get_stb(self):
        if self.show_tb:
            return self.stb
        else:
            return self.stb[-1:]

    def get_text(self):
        return self.text

    def show_last(self):
        print(self.text)



import pytest
from ..util import IMPLICITLINEEND, ENDMARKER, REFERENCE


@pytest.mark.parametrize("string", ("&m1", "&m2.T", "&m3.phi", "&m4.p1.o.q",))
def test_reference(tokenizer, string):
    assert list(tokenizer.tokenize(string)) == [
        REFERENCE(1, 1, string),
        IMPLICITLINEEND(1, 1 + len(string)),
        ENDMARKER(2),
    ]



import pytest
from ....util import dedent_multiline


@pytest.fixture
def compiler(compiler):
    element = lambda a=None, b=None, c=None, d=None: None
    compiler.spec.register_element(
        "fake_element1", {"setter": element, "getter": element}
    )
    compiler.spec.register_element(
        "fake_element2", {"setter": element, "getter": element}
    )
    compiler.spec.constants = {"A": 1, "B": "2", "C": object, "D": lambda: "hi"}
    compiler.spec.keywords = {"E", "F", "G", "H"}

    return compiler


@pytest.mark.parametrize(
    "script,graph_definition",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_element1 myelement1
                fake_element2 myelement2 myelement1.a
                """
            ),
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "['\\n']"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        type "ELEMENT"
                        token "'fake_element1'"
                        name_token "'myelement1'"
                        extra_tokens "[' ']"
                    ]
                    node [
                        id 2
                        label ".script.1"
                        type "ELEMENT"
                        token "'fake_element2'"
                        name_token "'myelement2'"
                        extra_tokens "[' ', ' ']"
                    ]
                    node [
                        id 3
                        label ".script.1.0"
                        token "'myelement1.a'"
                        type "PARAMETER"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 0
                        type "ARGUMENT"
                        order "1"
                    ]
                    edge [
                        source 3
                        target 2
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 1
                        target 3
                        type "DEPENDENCY"
                    ]
                ]
                """
            ),
            id="parameter between two elements",
        ),
    ),
)
def test_parameter(assert_graphs_match, script, graph_definition):
    assert_graphs_match(script, graph_definition)


@pytest.mark.parametrize(
    "script,graph_definition",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_element1 myelement1
                fake_element2 myelement2 &myelement1.a
                """
            ),
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "['\\n']"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        type "ELEMENT"
                        token "'fake_element1'"
                        name_token "'myelement1'"
                        extra_tokens "[' ']"
                    ]
                    node [
                        id 2
                        label ".script.1"
                        type "ELEMENT"
                        token "'fake_element2'"
                        name_token "'myelement2'"
                        extra_tokens "[' ', ' ']"
                    ]
                    node [
                        id 3
                        label ".script.1.0"
                        token "'&myelement1.a'"
                        type "PARAMETER_REFERENCE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 0
                        type "ARGUMENT"
                        order "1"
                    ]
                    edge [
                        source 3
                        target 2
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 1
                        target 3
                        type "DEPENDENCY"
                    ]
                ]
                """
            ),
            id="parameter reference between two elements",
        ),
    ),
)
def test_parameter_reference(assert_graphs_match, script, graph_definition):
    assert_graphs_match(script, graph_definition)



"""CLI fixtures.

See the `Click testing documentation <https://click.palletsprojects.com/en/7.x/testing/>`__.
"""

from tempfile import NamedTemporaryFile
import pytest
from click.testing import CliRunner


SCRIPT = """
laser l0 P=1
space s1 l0.p1 itm.p1
mirror itm R=0.99 T=0.01 Rc=-0.6
space scav itm.p2 etm.p1 L=1
mirror etm R=&itm.R T=&itm.T phi=&itm.phi Rc=0.6
cav cav1 itm.p2.o
pd pdcav itm.p1.o
"""


@pytest.fixture
def cli():
    """CLI runner for use in tests."""
    return CliRunner()


@pytest.fixture
def isolated_cli(cli):
    """CLI runner for use in tests, with isolated working directory."""
    with cli.isolated_filesystem():
        yield cli


@pytest.fixture
def input_file(tmp_path):
    """A temporary file containing valid kat script."""
    tmpfile = tmp_path / "kat.tmp"
    with open(tmpfile, "w") as fobj:
        fobj.write(SCRIPT)
    yield str(tmpfile)



"""Contains actions which can be performed on a model."""

from finesse.analysis.axes import (
    noxaxis,
    xaxis,
    x2axis,
    x3axis,
)

from finesse.analysis.beamtrace import (
    beam_trace,
    abcd,
)

from .actions import Action

__all__ = (
    "noxaxis",
    "xaxis",
    "x2axis",
    "x3axis",
    "beam_trace",
    "abcd",
    "Action"
)



import finesse
import pytest
from finesse.detectors import KnmDetector


@pytest.fixture()
def model_cavity():
    IFO = finesse.Model()
    IFO.parse(
        """
        l L0 P=3
        s s0 L0.p1 ITM.p1
        m ITM T=7m L=37.5u Rc=-5580
        s LY ITM.p2 ETM.p1 L=10k
        m ETM T=6u L=37.5u Rc=5580
        var dummy 6000
        cav ARM ITM.p2
        """
    )
    return IFO.deepcopy()


@pytest.mark.parametrize("mirror", ("ETM", "ITM"))
@pytest.mark.parametrize("direction", ("x", "y"))
@pytest.mark.parametrize("node1", (1, 2))
@pytest.mark.parametrize("node2", (1, 2))
@pytest.mark.parametrize("dummy", (True, False))
def test_ABCDs_xaxis_reset(model_cavity, mirror, direction, node1, node2, dummy):
    if dummy:
        # use dummy variable to scan RoCs instead
        model_cavity.ETM.Rcx = model_cavity.dummy.value.ref
        model_cavity.parse("xaxis(dummy.value, lin, 5000, 9000, 1)")
    else:
        model_cavity.parse("xaxis(ETM.Rcx, lin, 5000, 9000, 1)")

    model_cavity.ETM.Rcy = model_cavity.ETM.Rcx.ref
    model_cavity.ITM.Rcx = -1 * model_cavity.ETM.Rcx.ref
    model_cavity.ITM.Rcy = -1 * model_cavity.ETM.Rcx.ref

    m1 = model_cavity.elements[mirror].ABCD(node1, node2, direction=direction).copy()
    model_cavity.run()
    m2 = model_cavity.elements[mirror].ABCD(node1, node2, direction=direction).copy()
    assert (m1 == m2).all()


def test_cvalues_update_abcd(model_cavity):
    IFO = model_cavity
    IFO.ETM.Rcy = IFO.ETM.Rcx.ref
    IFO.ITM.Rcx = -1 * IFO.ETM.Rcx.ref
    IFO.ITM.Rcy = -1 * IFO.ETM.Rcx.ref
    IFO.parse("xaxis(ETM.Rcx, lin, 5580, 9000, 1)")
    IFO.add_matched_gauss(IFO.L0.p1.o, "gL0")
    IFO.add(KnmDetector("k00", IFO.ITM, "11", 0, 0, 0, 0))

    out1 = IFO.run()
    out2 = IFO.run()
    assert (out1["k00"] == out2["k00"]).all()


def test_cvalues_update_abcd_dummy_var(model_cavity):
    IFO = model_cavity
    IFO.ETM.Rcx = IFO.dummy.value.ref
    IFO.ETM.Rcy = IFO.dummy.value.ref
    IFO.ITM.Rcx = -1 * IFO.dummy.value.ref
    IFO.ITM.Rcy = -1 * IFO.dummy.value.ref
    IFO.parse("xaxis(dummy.value, lin, 5580, 9000, 1)")
    IFO.add_matched_gauss(IFO.L0.p1.o, "gL0")
    IFO.add(KnmDetector("k00", IFO.ITM, "11", 0, 0, 0, 0))

    out1 = IFO.run()
    out2 = IFO.run()
    assert (out1["k00"] == out2["k00"]).all()



"""
Space-type objects representing physical distances between components.
"""

import logging
import numpy as np
import types
from finesse.parameter import float_parameter
from finesse.components.general import (
    Connector,
    InteractionType,
    borrows_nodes,
    DOFDefinition,
)
from finesse.components.workspace import ConnectionSetting
from finesse.components.node import NodeDirection, NodeType, Port


LOGGER = logging.getLogger(__name__)


@borrows_nodes()
@float_parameter(
    "L",
    "Length",
    validate="_check_L",
    units="m",
    is_geometric=True,
)
@float_parameter(
    "nr",
    "Refractive index",
    validate="_check_nr",
    is_geometric=True,
    changeable_during_simulation=False,
)
@float_parameter("gouy_x", "Gouy phase (x)", units="Degrees")
@float_parameter("gouy_y", "Gouy phase (y)", units="Degrees")
class Space(Connector):
    """Represents a space between two components in the interferometer configuration, with a given \
    length and index of refraction.

    Parameters
    ----------
    name : str, optional
        Name of newly created space.

    portA, portB : :class:`.Port`, optional
        Ports to connect.

    L : float, optional
        Geometric length of newly created :class:`.Space` instance; defaults to 0.

    nr : float, optional
        Index of refraction of newly created :class:`.Space` instance; defaults to 1.0.
    """

    def __init__(self, name=None, portA=None, portB=None, L=0, nr=1.0, gouy_x=0, gouy_y=0):
        if (portA is None) != (portB is None):
            LOGGER.warn(
                "Can't construct a space with only one port connected; ignoring ports."
            )
            portA = None
            portB = None

        if portA is not None and not isinstance(portA, Port):
            raise Exception("PortA is not a Port")
        if portA is not None and not isinstance(portB, Port):
            raise Exception("PortB is not a Port")

        if portA is not None and portA.type != NodeType.OPTICAL:
            raise Exception("PortA is not an optical port")
        if portB is not None and portB.type != NodeType.OPTICAL:
            raise Exception("PortB is not an optical port")

        if portA is not None and portB is not None:
            if portA.component._model is not portB.component._model:
                raise Exception("Port A and B are not part of the same model")

        if name is None:
            if portA is not None and portB is not None:
                compA = portA.component.name
                compB = portB.component.name
                name = f"{compA}_{portA.name}__{compB}_{portB.name}"
            else:
                raise ValueError(
                    "Cannot create an unconnected space without providing a name"
                )

        super().__init__(name)

        self.__portA = portA
        self.__portB = portB
        self._add_to_model_namespace = True
        self._namespace = ".spaces"
        self.L = L
        self.nr = nr
        self.gouy_x.is_tunable = True
        self.gouy_y.is_tunable = True
        self.gouy_x = gouy_x
        self.gouy_y = gouy_y
        
        self._add_port("p1", NodeType.OPTICAL)
        self._add_port("p2", NodeType.OPTICAL)

        # Phase modulation input
        self._add_port("phs", NodeType.ELECTRICAL)
        self.phs._add_node("i", NodeDirection.INPUT)
        # Amplitude modulation input
        self._add_port("amp", NodeType.ELECTRICAL)
        self.amp._add_node("i", NodeDirection.INPUT)
        # strain input
        self._add_port("h", NodeType.ELECTRICAL)
        self.h._add_node("i", NodeDirection.INPUT)

        if portA is not None and portB is not None:
            self.connect(portA, portB)
        self.__changing_check = set((self.L, self.nr))

        # Define typical degrees of freedom for this component
        self.dofs = types.SimpleNamespace()
        # Strain doesn't have a DC term really in Finesse
        # changing a space length won't generate a signal
        self.dofs.h = DOFDefinition(None, self.h.i, 1)

    @property
    def portA(self):
        return self.__portA

    @property
    def portB(self):
        return self.__portB

    def connect(self, portA, portB):
        """
        Sets the ports of this `Space`.

        Parameters
        ----------
        portA : :class:`.Port`, optional
            Port to connect

        portB : :class:`.Port`, optional
            Port to connect
        """
        if portA.is_connected:
            raise Exception(f"Port {portA} has already been connected to")
        if portB.is_connected:
            raise Exception(f"Port {portB} has already been connected to")

        # From the Space's perspective the input and output
        # nodes are swapped around for its ports
        self.p1._add_node("i", None, portA.o)
        self.p1._add_node("o", None, portA.i)
        self.p2._add_node("i", None, portB.o)
        self.p2._add_node("o", None, portB.i)

        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling("SIGPHS_P1o", self.phs.i, self.p1.o)
        self._register_node_coupling("SIGPHS_P2o", self.phs.i, self.p2.o)
        self._register_node_coupling("SIGAMP_P1o", self.amp.i, self.p1.o)
        self._register_node_coupling("SIGAMP_P2o", self.amp.i, self.p2.o)
        self._register_node_coupling("H_P1o", self.h.i, self.p1.o)
        self._register_node_coupling("H_P2o", self.h.i, self.p2.o)

    def _get_workspace(self, sim):
        from finesse.components.modal.space import (
            space_carrier_fill,
            space_signal_fill,
            space_set_gouy,
            SpaceWorkspace,
        )

        _, is_changing = self._eval_parameters()

        carrier_refill = sim.carrier.any_frequencies_changing and (float(self.L.value) > 0 or self.L.is_changing)
        carrier_refill |= len(is_changing)
        carrier_refill |= sim.trace_forest.contains_space(self)

        ws = SpaceWorkspace(self, sim, False)
        ws.set_gouy_function(space_set_gouy)
        # Set the fill function for this simulation
        ws.carrier.add_fill_function(space_carrier_fill, carrier_refill)
        ws.carrier.connection_settings["P1i_P2o"] = ConnectionSetting.DIAGONAL
        ws.carrier.connection_settings["P2i_P1o"] = ConnectionSetting.DIAGONAL

        if sim.signal:
            signal_refill = sim.signal.any_frequencies_changing and (float(self.L.value) > 0 or self.L.is_changing)
            signal_refill |= len(is_changing)
            signal_refill |= sim.trace_forest.contains_space(self)
            
            ws.signal.add_fill_function(space_signal_fill, signal_refill)
            
            ws.signal.connection_settings["P1i_P2o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["P2i_P1o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["SIGPHS_P1o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["SIGPHS_P2o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["SIGAMP_P1o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["SIGAMP_P2o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["H_P1o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["H_P2o"] = ConnectionSetting.DIAGONAL

        # Initialise the ABCD matrix memory-views
        if sim.is_modal:
            ws.abcd = self.ABCD(self.p1.i, self.p2.o, "x", copy=False)

        return ws

    def _check_L(self, value):
        if value < 0:
            raise ValueError("Length of a space must not be negative.")

        return value

    def _check_nr(self, value):
        if value < 1:
            raise ValueError("Index of refraction must be >= 1")

        return value

    def _resymbolise_ABCDs(self):
        L = self.L.ref
        nr = self.nr.ref

        M_sym = np.array([[1.0, L / nr], [0.0, 1.0]])

        # Matrices same in both propagations and both planes so
        # only need one register call for all couplings here
        self.register_abcd_matrix(
            M_sym, (self.p1.i, self.p2.o), (self.p2.i, self.p1.o),
        )

    @property
    def abcd(self):
        """Numeric ABCD matrix.

        Equivalent to any of ``space.ABCD(1, 2, "x")``, ``space.ABCD(2, 1, "x")``,
        ``space.ABCD(1, 2, "y")``, ``space.ABCD(2, 1, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix (read-only).
        """
        return self.ABCD(1, 2, "x")

    def ABCD(
        self,
        from_node,
        to_node,
        direction="x",
        symbolic=False,
        copy=True,
        retboth=False,
    ):
        r"""Returns the ABCD matrix of the space for the specified coupling.

        .. _fig_abcd_space_transmission:
        .. figure:: ../images/abcd_spacet.*
            :align: center

        This is given by,

        .. math::
            M = \begin{pmatrix}
                    1 & \frac{L}{n_r} \\
                    0 & 1
                \end{pmatrix},

        where :math:`L` is the length of the space and :math:`n_r` is
        the index of refraction.

        See :meth:`.Connector.ABCD` for descriptions of parameters, return values and possible
        exceptions.
        """
        return super().ABCD(from_node, to_node, direction, symbolic, copy, retboth)



import numpy as np

from finesse.detectors.general import Detector
from finesse.detectors.compute.gaussian import BPDetectorWorkspace, BeamProperty


# Map of beam property keywords to enum fields.
BP_KEYWORDS = {
    "w": BeamProperty.SIZE,
    "w0": BeamProperty.WAISTSIZE,
    "z": BeamProperty.DISTANCE,
    "zr": BeamProperty.RAYLEIGH,
    "gouy": BeamProperty.GOUY,
    "div": BeamProperty.DIVERGENCE,
    "rc": BeamProperty.ROC,
    "s": BeamProperty.DEFOCUS,
    "q": BeamProperty.Q,
}


class BeamPropertyDetector(Detector):
    r"""Probe for detecting the properties of a beam at a given node.

    The valid values for `prop` are:

      * ``"w"``: beam size at `node` [metres],
      * ``"w0"``: waist size as measured at `node` [metres],
      * ``"z"``: distance to the waist from `node` [metres],
      * ``"zr"``: the Rayleigh range [metres],
      * ``"gouy"``: the Gouy phase of the beam at `node` [radians],
      * ``"div"``: divergence angle of the beam at `node`  [radians],
      * ``"rc"``: radius of curvature of wavefront at `node` [metres],
      * ``"s"``: curvature of wavefront at `node` [1 / metres],
      * ``"q"``: beam parameter at `node`.

    .. note::

        The ``"gouy"`` target property here detects the Gouy phase as derived
        from the beam parameter :math:`q` at the specified node, i.e:

        .. math::

            \psi = \arctan{\left(\frac{\myRe{q}}{\myIm{q}}\right)}.

        It does **not** compute any Gouy phase accumulation. Use :class:`.Gouy`
        to detect the accumulated Gouy phase over a path.

    Parameters
    ----------
    name : str
        Name of newly created detector.

    node : :class:`.OpticalNode`
        Node to read output from.

    prop : str or :class:`.BeamProperty`
        The property of the beam to detect. See above for options.

    direction : str, optional; default: 'x'
        Plane to detect in - 'x' for tangential, 'y' for sagittal.

    q_as_bp : bool, optional; default: False
        If detecting q, should the detector output return :class:`.BeamParam`
        object instead of just a complex number.
    """

    def __init__(self, name, node, prop, direction="x", q_as_bp=False):
        if isinstance(prop, str):
            if prop.casefold() not in BP_KEYWORDS:
                raise ValueError(
                    f"Unrecognised property: {prop}, expected "
                    f"one of: {list(BP_KEYWORDS.keys())}"
                )
            prop = BP_KEYWORDS[prop.casefold()]

        if prop == BeamProperty.Q:
            if q_as_bp:
                dtype = object
            else:
                dtype = np.complex128
            units = ""
        else:
            dtype = np.float64
            if prop == BeamProperty.GOUY or prop == BeamProperty.DIVERGENCE:
                units = "radians"
            elif prop == BeamProperty.DEFOCUS:
                units = "1/m"
            else:
                units = "m"

        property_to_label = {
            BeamProperty.SIZE: "Beam size",
            BeamProperty.WAISTSIZE: "Beam waist-size",
            BeamProperty.DISTANCE: "Distance to beam waist",
            BeamProperty.RAYLEIGH: "Rayleigh range",
            BeamProperty.GOUY: "Gouy phase",
            BeamProperty.DIVERGENCE: "Divergence angle",
            BeamProperty.ROC: "Beam radius of curvature",
            BeamProperty.DEFOCUS: "Beam defocus",
            BeamProperty.Q: "Beam parameter",
        }
        label = property_to_label[prop]

        Detector.__init__(self, name, node, dtype=dtype, unit=units, label=label)
        self.__prop = prop
        self.direction = direction

        self.q_as_bp = q_as_bp

    @property
    def prop(self):
        return self.__prop

    @property
    def needs_fields(self):
        return False

    @property
    def needs_trace(self):
        return True

    @property
    def detecting(self):
        """The property of the beam which is being detected.

        :getter: Returns the detected property (read-only).
        """
        return self.__prop

    def _get_workspace(self, sim):
        ws = BPDetectorWorkspace(self, sim)
        ws.q_as_bp = self.q_as_bp
        return ws



"""Test cases for selecting modes to include in a model."""

from finesse import Model
from hypothesis import given, assume
from hypothesis.strategies import one_of, composite, integers, complex_numbers, floats
import pytest


@pytest.fixture(scope="module")
def model():
    """Fixture with module scope for quick fuzzing tests."""
    return Model()


@composite
def non_whole_floats(draw):
    value = draw(floats())
    if value.is_integer():
        assume(False)
    return value


@given(maxtem=integers(max_value=-1))
def test_negative_integer_maxtem_is_invalid(model, maxtem):
    """Test that maxtem cannot be negative integer."""
    with pytest.raises(ValueError):
        model.select_modes(maxtem=maxtem)


@given(maxtem=one_of(non_whole_floats(), complex_numbers()))
def test_non_integer_maxtem_is_invalid(model, maxtem):
    """Test that maxtem cannot be non-integer."""
    with pytest.raises((ValueError, TypeError)):
        model.select_modes(maxtem=maxtem)



"""
Plotting tools for Finesse, providing convenient style templates for
:mod:`matplotlib.pyplot` and functions for quick visualisation of detector
and/or probe outputs.
"""

from .style import list_styles, use, context
from .plot import Plotter, bode
from .graph import plot_graph, plot_nx_graph, plot_graphviz
from .tools import init


__all__ = (
    "Plotter",
    "init",
    "list_styles",
    "use",
    "context",
    "bode",
    "plot_graph",
    "plot_nx_graph",
    "plot_graphviz",
)



"""Tests on propagate_beam tracing tools function, using a model of the aLIGO PRX + XARM
cavities."""

from functools import reduce
import numpy as np
from numpy.testing import assert_allclose
import pytest
from finesse import Model
from finesse.components import Gauss
from finesse.gaussian import BeamParam
from finesse.symbols import evaluate
from finesse.utilities import pairwise


@pytest.fixture()
def aligo_prx_model():
    IFO = Model()
    IFO.parse(
        """
    variable nsilica 1.45
    variable Mloss 30u
    variable Larm 3994

    ###   laser
    ###########################################################################
    # Laser and input optics
    laser L0 P=125
    link(L0, PRMAR)

    ###########################################################################
    ###   PRC
    ###########################################################################
    m PRMAR R=0 L=40u xbeta=&PRM.xbeta ybeta=&PRM.ybeta phi=&PRM.phi
    s sPRMsub1 PRMAR.p2 PRM.p1 L=0.0737 nr=&nsilica
    m PRM T=0.03 L=8.5u Rc=11.009
    s lp1 PRM.p2 PR2.p1 L=16.6107
    bs PR2 T=250u L=&Mloss alpha=-0.79 Rc=-4.545
    s lp2 PR2.p2 PR3.p1 L=16.1647
    bs PR3 T=0 L=&Mloss alpha=0.615 Rc=36.027
    s lp3 PR3.p2 BS.p1 L=19.5381

    ###########################################################################
    ###   BS
    ###########################################################################
    bs BS R=0.5 L=&Mloss alpha=45
    s BSsub1 BS.p3 BSAR1.p1 L=0.0687 nr=&nsilica
    s BSsub2 BS.p4 BSAR2.p2 L=0.0687 nr=&nsilica
    bs BSAR1 L=50u R=0 alpha=-29.195
    bs BSAR2 L=50u R=0 alpha=29.195

    ###########################################################################
    ###   Xarm
    ###########################################################################
    # Distance from beam splitter to X arm input mirror
    s lx1 BSAR1.p3 ITMXlens.p1 L=4.993
    lens ITMXlens f=34500
    s lx2 ITMXlens.p2 ITMXAR.p1
    m ITMXAR R=0 L=20u xbeta=&ITMX.xbeta ybeta=&ITMX.ybeta phi=&ITMX.phi
    s ITMXsub ITMXAR.p2 ITMX.p1 L=0.2 nr=&nsilica
    m ITMX T=0.014 L=&Mloss Rc=-1934

    s LX ITMX.p2 ETMX.p1 L=&Larm

    m ETMX T=5u L=&Mloss Rc=2245
    s ETMXsub ETMX.p2 ETMXAR.p1 L=0.2 nr=&nsilica
    m ETMXAR 0 500u xbeta=&ETMX.xbeta ybeta=&ETMX.ybeta phi=&ETMX.phi

    ################
    ### Cavities ###
    ################

    cav cavXARM ITMX.p2
    cav cavPRX PRM.p2 ITMX.p1.i
    """
    )

    return IFO


@pytest.mark.parametrize(
    "start_end",
    [
        ("PRM.p2.o", "ITMX.p1.i"),
        ("PRM.p2", "ETMX.p1"),
        ("L0.p1", "ITMXlens.p2.o"),
        ("PR3.p1.i", "BS.p1"),
    ],
)
@pytest.mark.parametrize("use_path", (True, False))
def test_start_end_nodes(aligo_prx_model: Model, start_end, use_path):
    """Test path creation in propagate_beam via start and end node checking."""
    IFO = aligo_prx_model

    start, end = start_end
    if use_path:
        beam = IFO.propagate_beam(path=IFO.path(start, end))
    else:
        beam = IFO.propagate_beam(start, end)

    # Expect start ports to default to output node...
    if start.count(".") == 1:
        start += ".o"
    # ... and end ports to input node
    if end.count(".") == 1:
        end += ".i"

    start = IFO.reduce_get_attr(start)
    end = IFO.reduce_get_attr(end)

    assert beam.start_node is start
    assert beam.end_node is end


@pytest.mark.parametrize("use_path", (True, False))
def test_same_start_end_node_raises_exception(aligo_prx_model: Model, use_path):
    """Test that attempting to use same start and end node for propagate_beam call
    raises an exception."""
    IFO = aligo_prx_model

    with pytest.raises(ValueError):
        if use_path:
            IFO.propagate_beam(path=IFO.path(IFO.PRM.p2.o, IFO.PRM.p2.o))
        else:
            IFO.propagate_beam(IFO.PRM.p2.o, IFO.PRM.p2.o)


@pytest.mark.parametrize(
    "start_end",
    [("PRM.p2.o", "ITMX.p1.i"), ("ITMX.p2.i", "PRM.p2"), ("L0.p1", "ITMXlens.p2.o"),],
)
@pytest.mark.parametrize(
    "enable_only", [("cavXARM", "cavPRX"), ("cavXARM",), ("cavPRX",)]
)
@pytest.mark.parametrize("direction", ("x", "y"))
def test_default_qin(aligo_prx_model: Model, start_end, enable_only, direction):
    """Test that specifying no input q uses the correct beam parameter for given start
    node."""
    IFO = aligo_prx_model

    start, end = start_end
    beam = IFO.propagate_beam(start, end, enable_only=enable_only, direction=direction)
    trace = IFO.beam_trace(enable_only=enable_only)

    # Expect start ports to default to output node
    if start.count(".") == 1:
        start += ".o"

    start = IFO.reduce_get_attr(start)

    assert getattr(trace[start], f"q{direction}") == beam.q(beam.start_node)


@pytest.mark.parametrize(
    "start_end",
    [
        ("PRM.p2.o", "ITMX.p1.i"),
        ("ITMX.p2.i", "PRM.p2"),
        ("L0.p1", "ITMXlens.p2.o"),
        ("PR3.p2.o", "BS.p1.i"),
    ],
)
@pytest.mark.parametrize("direction", ("x", "y"))
@pytest.mark.parametrize("symbolic", (True, False))
def test_composite_abcd(aligo_prx_model: Model, start_end, direction, symbolic):
    """Test 'composite' ABCD matrices computed by propagate_beam."""
    IFO = aligo_prx_model

    start, end = start_end
    beam = IFO.propagate_beam(start, end, direction=direction, symbolic=symbolic)

    # Collect the individual ABCD matrices of each component
    abcds = []
    for n1, n2 in pairwise(beam.nodes):
        abcd = IFO.ABCD(n1, n2, direction=direction, symbolic=symbolic).M
        abcds.append(evaluate(abcd))

    # And compute the full matrix on these
    expect_abcd = reduce(np.dot, reversed(abcds))

    assert_allclose(evaluate(beam.full_ABCD), expect_abcd)


@pytest.mark.parametrize(
    "start_end",
    [
        ("PRM.p2.o", "ITMX.p1.i"),
        ("ITMX.p1", "PRM.p2"),
        ("L0.p1", "ITMXlens.p2.o"),
        ("PR3.p2.o", "BS.p1.i"),
    ],
)
@pytest.mark.parametrize("direction", ("x", "y"))
@pytest.mark.parametrize("symbolic", (True, False))
def test_accumulated_gouy_phases(
    aligo_prx_model: Model, start_end, direction, symbolic
):
    """Test Gouy phases accumulated over specified paths computed via propagate_beam,
    against result from simulation using Gouy detector."""
    IFO = aligo_prx_model

    start, end = start_end
    beam = IFO.propagate_beam(start, end, direction=direction, symbolic=symbolic)

    startn = start.replace(".", "_")
    endn = end.replace(".", "_")
    IFO.parse(
        f"gouy acc_gouy_{startn}_{endn} from_node={start} to_node={end} direction={direction}"
    )
    out = IFO.run()

    assert evaluate(beam.total_acc_gouy) == pytest.approx(
        out[f"acc_gouy_{startn}_{endn}"]
    )


@pytest.mark.parametrize(
    "cutoff", ["PRM", "PR2", "PR3", "BS", "BSAR1", "ITMXlens", "ITMX"]
)
@pytest.mark.parametrize("direction", ("x", "y"))
def test_accumulated_gouy_phases__cut_off(
    aligo_prx_model: Model, cutoff, direction,
):
    """Test Gouy phases accumulated over specified paths computed via propagate_beam, up
    to a specific cut-off component, against result from simulation using Gouy
    detector."""
    IFO = aligo_prx_model

    beam = IFO.propagate_beam("PRM.p2", "ITMX.p1", direction=direction)

    cutoff_node = {
        "PRM": "PRM.p2.o",
        "PR2": "PR2.p1.i",
        "PR3": "PR3.p1.i",
        "BS": "BS.p1.i",
        "BSAR1": "BSAR1.p1.i",
        "ITMXlens": "ITMXlens.p1.i",
        "ITMX": "ITMX.p1.i",
    }
    IFO.parse(
        f"gouy acc_gouy_{cutoff} from_node=PRM.p2 to_node={cutoff_node[cutoff]} direction={direction}"
    )
    out = IFO.run()

    assert beam.acc_gouy_up_to(cutoff) == pytest.approx(out[f"acc_gouy_{cutoff}"])


@pytest.mark.parametrize(
    "start_end",
    [
        ("PRM.p2.o", "ITMX.p1.i"),
        ("ITMX.p1", "PRM.p2"),
        ("L0.p1", "ITMXlens.p2.o"),
        ("PR3.p2.o", "BS.p1.i"),
    ],
)
@pytest.mark.parametrize("direction", ("x", "y"))
def test_all_node_qs__same_cav(aligo_prx_model: Model, start_end, direction):
    """Test the beam parameters at all nodes of the propagation solution are correct,
    where the start and end nodes of the propagate_beam call are in the same cavity
    path."""
    IFO = aligo_prx_model

    start, end = start_end
    beam = IFO.propagate_beam(start, end, direction=direction)
    qs = beam.qs

    trace = IFO.beam_trace()

    for node, q in qs.items():
        assert q == getattr(trace[node], f"q{direction}")


@pytest.mark.parametrize(
    "start_end", [("PRM.p2.o", "ETMX.p1.i"), ("ITMX.p2.i", "PRM.p2"),],
)
@pytest.mark.parametrize("direction", ("x", "y"))
def test_all_node_qs__diff_cav(aligo_prx_model: Model, start_end, direction):
    """Test the beam parameters at all nodes of the propagation solution are correct,
    where the start and end nodes of the propagate_beam call are in different cavity
    paths --- only the cavity associated with start node should then be enabled."""
    IFO = aligo_prx_model

    start, end = start_end
    beam = IFO.propagate_beam(start, end, direction=direction)
    qs = beam.qs

    start_node = IFO.reduce_get_attr(start)
    # Get the cavity associated with the start node and only
    # enable this for the beam trace
    cav = None
    for cav in IFO.cavities:
        if start_node in cav.path.nodes_only:
            cav = cav
            break

    assert cav is not None

    trace = IFO.beam_trace(enable_only=cav)

    for node, q in qs.items():
        assert q == getattr(trace[node], f"q{direction}")


@pytest.mark.parametrize("qin", [7 + 5j, 6.52 + 7.1j, 5.31j])
@pytest.mark.parametrize("direction", ("x", "y"))
def test_all_node_qs__custom_qin(aligo_prx_model: Model, qin, direction):
    """Test the beam parameters at all nodes of the propagation solution are correct,
    where a custom input beam parameter is set."""
    IFO = aligo_prx_model

    beam = IFO.propagate_beam("PRM.p2", "ETMX.p1", q_in=qin, direction=direction)
    qs = beam.qs

    # Set a Gauss at PRM.p2.o using custom qin
    IFO.add(Gauss("gPRM2o", IFO.PRM.p2.o, q=qin))
    trace = IFO.beam_trace(enable_only="gPRM2o")

    for node, q in qs.items():
        assert q == getattr(trace[node], f"q{direction}")


@pytest.mark.parametrize(
    "start_end",
    [
        ("PRM.p2.o", "ITMX.p1.i"),
        ("ITMX.p1", "PRM.p2"),
        ("L0.p1", "ITMXlens.p2.o"),
        ("PR3.p2.o", "BS.p1.i"),
    ],
)
@pytest.mark.parametrize("symbolic", (True, False))
def test_astigmatic_propagation_overlaps(aligo_prx_model: Model, start_end, symbolic):
    """Test that overlaps between tangential and sagittal plane beam parameters are
    correct when using propagate_beam_astig function."""
    IFO = aligo_prx_model

    start, end = start_end
    beam = IFO.propagate_beam_astig(start, end, symbolic=symbolic)
    overlaps = beam.overlaps

    trace = IFO.beam_trace()

    for node, Ov in overlaps.items():
        assert evaluate(Ov) == pytest.approx(BeamParam.overlap(*trace[node]))



import pytest
from finesse.script.containers import KatFile, KatElement, KatKwarg
from finesse.script.exceptions import KatScriptError
from ...util import dedent_multiline
from ..util import NAME, NUMBER, SPACE, EQUALS


@pytest.fixture
def three_components():
    """Kat file representing three components."""
    return KatFile(
        dedent_multiline(
            """
            fake_component mycomponent1 A=1 B=2
            fake_component mycomponent2 A=3 B=4
            fake_component mycomponent3 A=5 B=6

            fake_component mycomponent4 A=7 B=8
            fake_component mycomponent5 A=9 B=10
            """
        )
    )


@pytest.mark.parametrize(
    "error_items,error",
    (
        # Only lines 1 and 2 are shown because the error is on the first line.
        (
            [[NAME(1, 16, "mycomponent1")]],
            (
                "line 1: __error_msg__\n"
                "-->1: fake_component mycomponent1 A=1 B=2\n"
                "                     ^^^^^^^^^^^^\n"
                "   2: fake_component mycomponent2 A=3 B=4"
            ),
        ),
        # Lines 1-3 shown.
        (
            [[NAME(2, 16, "mycomponent2")]],
            (
                "line 2: __error_msg__\n"
                "   1: fake_component mycomponent1 A=1 B=2\n"
                "-->2: fake_component mycomponent2 A=3 B=4\n"
                "                     ^^^^^^^^^^^^\n"
                "   3: fake_component mycomponent3 A=5 B=6"
            ),
        ),
        # Lines 2-4 shown.
        (
            [[NAME(3, 16, "mycomponent3")]],
            (
                "line 3: __error_msg__\n"
                "   2: fake_component mycomponent2 A=3 B=4\n"
                "-->3: fake_component mycomponent3 A=5 B=6\n"
                "                     ^^^^^^^^^^^^\n"
                "   4: "
            ),
        ),
        # Lines 4-6 shown.
        (
            [[NAME(5, 16, "mycomponent4")]],
            (
                "line 5: __error_msg__\n"
                "   4: \n"
                "-->5: fake_component mycomponent4 A=7 B=8\n"
                "                     ^^^^^^^^^^^^\n"
                "   6: fake_component mycomponent5 A=9 B=10"
            ),
        ),
        # Only lines 6 and 7 are shown because the error is on the last line.
        (
            [[NAME(6, 16, "mycomponent5")]],
            (
                "line 6: __error_msg__\n"
                "   5: fake_component mycomponent4 A=7 B=8\n"
                "-->6: fake_component mycomponent5 A=9 B=10\n"
                "                     ^^^^^^^^^^^^"
            ),
        ),
    ),
)
def test_token_error(three_components, error_items, error):
    exception = KatScriptError("__error_msg__", three_components, error_items)
    assert exception.message() == error


@pytest.mark.parametrize(
    "error_items,error",
    (
        pytest.param(
            [
                [
                    KatElement(
                        directive=NAME(1, 1, "fake_component"),
                        name=NAME(1, 16, "mycomponent1"),
                        arguments=[
                            KatKwarg(
                                key=NAME(1, 29, "A"),
                                value=NUMBER(1, 31, "1"),
                                extra=[EQUALS(1, 30)],
                            ),
                            KatKwarg(
                                key=NAME(1, 33, "B"),
                                value=NUMBER(1, 35, "2"),
                                extra=[EQUALS(1, 34)],
                            ),
                        ],
                        extra=[SPACE(1, 15), SPACE(1, 28)],
                    )
                ]
            ],
            (
                "line 1: __error_msg__\n"
                "-->1: fake_component mycomponent1 A=1 B=2\n"
                "      ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n"
                "   2: fake_component mycomponent2 A=3 B=4"
            ),
            id="error on first line",
        ),
        pytest.param(
            [
                [
                    KatElement(
                        directive=NAME(2, 1, "fake_component"),
                        name=NAME(2, 16, "mycomponent2"),
                        arguments=[
                            KatKwarg(
                                key=NAME(2, 29, "A"),
                                value=NUMBER(2, 31, "3"),
                                extra=[EQUALS(2, 30)],
                            ),
                            KatKwarg(
                                key=NAME(2, 33, "B"),
                                value=NUMBER(2, 35, "4"),
                                extra=[EQUALS(2, 34)],
                            ),
                        ],
                        extra=[SPACE(2, 15), SPACE(2, 28)],
                    ),
                ]
            ],
            (
                "line 2: __error_msg__\n"
                "   1: fake_component mycomponent1 A=1 B=2\n"
                "-->2: fake_component mycomponent2 A=3 B=4\n"
                "      ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n"
                "   3: fake_component mycomponent3 A=5 B=6"
            ),
            id="error on intermediate (second) line",
        ),
        pytest.param(
            [
                [
                    KatElement(
                        directive=NAME(3, 1, "fake_component"),
                        name=NAME(3, 16, "mycomponent3"),
                        arguments=[
                            KatKwarg(
                                key=NAME(3, 29, "A"),
                                value=NUMBER(3, 31, "5"),
                                extra=[EQUALS(3, 30)],
                            ),
                            KatKwarg(
                                key=NAME(3, 33, "B"),
                                value=NUMBER(3, 35, "6"),
                                extra=[EQUALS(3, 34)],
                            ),
                        ],
                        extra=[SPACE(3, 15), SPACE(3, 28)],
                    ),
                ]
            ],
            (
                "line 3: __error_msg__\n"
                "   2: fake_component mycomponent2 A=3 B=4\n"
                "-->3: fake_component mycomponent3 A=5 B=6\n"
                "      ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n"
                "   4: "
            ),
            id="error on intermediate (third) line",
        ),
        pytest.param(
            [
                [
                    KatElement(
                        directive=NAME(5, 1, "fake_component"),
                        name=NAME(5, 16, "mycomponent4"),
                        arguments=[
                            KatKwarg(
                                key=NAME(5, 29, "A"),
                                value=NUMBER(5, 31, "7"),
                                extra=[EQUALS(5, 30)],
                            ),
                            KatKwarg(
                                key=NAME(5, 33, "B"),
                                value=NUMBER(5, 35, "8"),
                                extra=[EQUALS(5, 34)],
                            ),
                        ],
                        extra=[SPACE(5, 15), SPACE(5, 28)],
                    ),
                ]
            ],
            (
                "line 5: __error_msg__\n"
                "   4: \n"
                "-->5: fake_component mycomponent4 A=7 B=8\n"
                "      ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n"
                "   6: fake_component mycomponent5 A=9 B=10"
            ),
            id="error on intermediate (fifth) line",
        ),
        pytest.param(
            [
                [
                    KatElement(
                        directive=NAME(6, 1, "fake_component"),
                        name=NAME(6, 16, "mycomponent5"),
                        arguments=[
                            KatKwarg(
                                key=NAME(6, 29, "A"),
                                value=NUMBER(6, 31, "9"),
                                extra=[EQUALS(6, 30)],
                            ),
                            KatKwarg(
                                key=NAME(6, 33, "B"),
                                value=NUMBER(6, 35, "10"),
                                extra=[EQUALS(6, 34)],
                            ),
                        ],
                        extra=[SPACE(6, 15), SPACE(6, 28)],
                    ),
                ]
            ],
            (
                "line 6: __error_msg__\n"
                "   5: fake_component mycomponent4 A=7 B=8\n"
                "-->6: fake_component mycomponent5 A=9 B=10\n"
                "      ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"
            ),
            id="error on last line",
        ),
    ),
)
def test_element_error(three_components, error_items, error):
    exception = KatScriptError("__error_msg__", three_components, error_items)
    assert exception.message() == error


@pytest.mark.parametrize(
    "error_items,error",
    (
        # Only lines 1 and 2 are shown because the error is on the first line.
        (
            [
                [
                    KatElement(
                        directive=NAME(1, 1, "fake_component"),
                        name=NAME(1, 16, "mycomponent1"),
                        arguments=[
                            KatKwarg(
                                key=NAME(1, 29, "A"),
                                value=NUMBER(1, 31, "1"),
                                extra=[EQUALS(1, 30)],
                            ),
                            KatKwarg(
                                key=NAME(1, 29, "A"),
                                value=NUMBER(1, 35, "2"),
                                extra=[EQUALS(1, 34)],
                            ),
                        ],
                        extra=[SPACE(1, 15), SPACE(1, 28)],
                    ),
                    NUMBER(1, 31, "1"),
                ]
            ],
            (
                "line 1: __error_msg__\n"
                "-->1: fake_component mycomponent1 A=1 B=2\n"
                "                                    ^\n"
                "   2: fake_component mycomponent2 A=3 B=4"
            ),
        ),
        # Lines 1-3 shown.
        (
            [
                [
                    KatElement(
                        directive=NAME(2, 1, "fake_component"),
                        name=NAME(2, 16, "mycomponent2"),
                        arguments=[
                            KatKwarg(
                                key=NAME(2, 29, "A"),
                                value=NUMBER(2, 31, "3"),
                                extra=[EQUALS(2, 30)],
                            ),
                            KatKwarg(
                                key=NAME(2, 33, "B"),
                                value=NUMBER(2, 35, "4"),
                                extra=[EQUALS(2, 34)],
                            ),
                        ],
                        extra=[SPACE(2, 15), SPACE(2, 28)],
                    ),
                    NUMBER(2, 35, "4"),
                ]
            ],
            (
                "line 2: __error_msg__\n"
                "   1: fake_component mycomponent1 A=1 B=2\n"
                "-->2: fake_component mycomponent2 A=3 B=4\n"
                "                                        ^\n"
                "   3: fake_component mycomponent3 A=5 B=6"
            ),
        ),
        # Lines 2-4 shown.
        (
            [
                [
                    KatElement(
                        directive=NAME(3, 1, "fake_component"),
                        name=NAME(3, 16, "mycomponent3"),
                        arguments=[
                            KatKwarg(
                                key=NAME(3, 29, "A"),
                                value=NUMBER(3, 31, "5"),
                                extra=[EQUALS(3, 30)],
                            ),
                            KatKwarg(
                                key=NAME(3, 29, "A"),
                                value=NUMBER(3, 35, "6"),
                                extra=[EQUALS(3, 34)],
                            ),
                        ],
                        extra=[SPACE(3, 15), SPACE(3, 28)],
                    ),
                    NAME(3, 29, "A"),
                ]
            ],
            (
                "line 3: __error_msg__\n"
                "   2: fake_component mycomponent2 A=3 B=4\n"
                "-->3: fake_component mycomponent3 A=5 B=6\n"
                "                                  ^\n"
                "   4: "
            ),
        ),
        # Lines 4-6 shown.
        (
            [
                [
                    KatElement(
                        directive=NAME(5, 1, "fake_component"),
                        name=NAME(5, 16, "mycomponent4"),
                        arguments=[
                            KatKwarg(
                                key=NAME(5, 29, "A"),
                                value=NUMBER(5, 31, "7"),
                                extra=[EQUALS(5, 30)],
                            ),
                            KatKwarg(
                                key=NAME(5, 33, "A"),
                                value=NUMBER(5, 35, "8"),
                                extra=[EQUALS(5, 34)],
                            ),
                        ],
                        extra=[SPACE(5, 15), SPACE(5, 28)],
                    ),
                    NAME(5, 29, "A"),
                ]
            ],
            (
                "line 5: __error_msg__\n"
                "   4: \n"
                "-->5: fake_component mycomponent4 A=7 B=8\n"
                "                                  ^\n"
                "   6: fake_component mycomponent5 A=9 B=10"
            ),
        ),
        # Only lines 5 and 6 are shown because the error is on the last line.
        (
            [
                [
                    KatElement(
                        directive=NAME(6, 1, "fake_component"),
                        name=NAME(6, 16, "mycomponent5"),
                        arguments=[
                            KatKwarg(
                                key=NAME(6, 29, "A"),
                                value=NUMBER(6, 31, "9"),
                                extra=[EQUALS(6, 30)],
                            ),
                            KatKwarg(
                                key=NAME(6, 29, "A"),
                                value=NUMBER(6, 35, "10"),
                                extra=[EQUALS(6, 34)],
                            ),
                        ],
                        extra=[SPACE(6, 15), SPACE(6, 28)],
                    ),
                    NAME(6, 29, "A"),
                ]
            ],
            (
                "line 6: __error_msg__\n"
                "   5: fake_component mycomponent4 A=7 B=8\n"
                "-->6: fake_component mycomponent5 A=9 B=10\n"
                "                                  ^"
            ),
        ),
    ),
)
def test_token_error_inside_element(three_components, error_items, error):
    exception = KatScriptError("__error_msg__", three_components, error_items)
    assert exception.message() == error


@pytest.mark.parametrize(
    "error_items,error",
    (
        # Errors on lines 1 and 3; lines 1-4 shown.
        (
            [[NUMBER(1, 31, "1")], [NUMBER(3, 31, "5")]],
            (
                "lines 1 and 3: __error_msg__\n"
                "-->1: fake_component mycomponent1 A=1 B=2\n"
                "                                    ^\n"
                "   2: fake_component mycomponent2 A=3 B=4\n"
                "-->3: fake_component mycomponent3 A=5 B=6\n"
                "                                    ^\n"
                "   4: "
            ),
        ),
        # Errors on lines 1 and 5; lines 1-2, 4-6 shown; 1 missing line message shown.
        (
            [[NUMBER(1, 31, "1")], [NUMBER(5, 31, "7")]],
            (
                "lines 1 and 5: __error_msg__\n"
                "-->1: fake_component mycomponent1 A=1 B=2\n"
                "                                    ^\n"
                "   2: fake_component mycomponent2 A=3 B=4\n"
                "      *** 1 skipped line ***\n"
                "   4: \n"
                "-->5: fake_component mycomponent4 A=7 B=8\n"
                "                                    ^\n"
                "   6: fake_component mycomponent5 A=9 B=10"
            ),
        ),
        # Errors on lines 1 and 6; lines 1-2, 5-6 shown; 2 missing lines message shown.
        (
            [[NUMBER(1, 31, "1")], [NUMBER(6, 31, "9")]],
            (
                "lines 1 and 6: __error_msg__\n"
                "-->1: fake_component mycomponent1 A=1 B=2\n"
                "                                    ^\n"
                "   2: fake_component mycomponent2 A=3 B=4\n"
                "      *** 2 skipped lines ***\n"
                "   5: fake_component mycomponent4 A=7 B=8\n"
                "-->6: fake_component mycomponent5 A=9 B=10\n"
                "                                    ^"
            ),
        ),
    ),
)
def test_errors_in_separate_locations(three_components, error_items, error):
    exception = KatScriptError("__error_msg__", three_components, error_items)
    assert exception.message() == error



import pytest
from finesse.script.spec import KatSpec

# The fixtures below have package scope because the tests in this directory should not try to
# modify the spec or compiler in any way. Tests that modify the spec or compiler should go
# elsewhere.


@pytest.fixture(scope="package")
def spec():
    return KatSpec()



"""The beam tracing library of Finesse.

This sub-package includes the core data structures and algorithms used for structuring
and performing beam traces (:mod:`.ctracer`, for developer reference mostly), tools for
executing propagations of beams outside of the context of a simulation (:mod:`.tools`)
and an optimiser which uses these tools (:mod:`.optimise`).
"""



import finesse.ligo
import pytest


def test_parse():
    model = finesse.Model()
    model.parse(finesse.ligo.aligo_katscript)


def test_run_aligo_noxaxis():
    model = finesse.Model()
    model.parse(finesse.ligo.aligo_katscript)
    model.run()


def test_SRC_PRC_length_symbols():
    model = finesse.Model()
    model.parse(finesse.ligo.aligo_katscript)
    path = model.path("PRM.p2.o", "ITMX.p1.i")
    LPRCX = sum(s.L * s.nr for s in path.spaces).eval()
    path = model.path("PRM.p2.o", "ITMY.p1.i")
    LPRCY = sum(s.L * s.nr for s in path.spaces).eval()
    assert (LPRCX + LPRCY) / 2 == model.lPRC.value.eval()

    path = model.path("SRM.p1.o", "ITMX.p1.i")
    LSRCX = sum(s.L * s.nr for s in path.spaces).eval()
    path = model.path("SRM.p1.o", "ITMY.p1.i")
    LSRCY = sum(s.L * s.nr for s in path.spaces).eval()
    assert (LSRCX + LSRCY) / 2 == model.lSRC.value.eval()


def test_make_model():
    finesse.ligo.make_aligo()


def test_locks():
    base = finesse.ligo.make_aligo()
    base.parse("run_locks()")
    base.run()


@pytest.mark.parametrize('L', (0.01, 0.05, 0.1))
def test_schnupp_length_experiment(L):
    """RF locks each arm and measures the optimal AS45 demodulation phase. From this
    back out the schnupp asymetry."""
    import numpy as np
    from finesse.ligo import make_aligo, DRFPMI_state
    from finesse.analysis.actions import Series, OptimiseRFReadoutPhaseDC
    base = finesse.ligo.make_aligo()
    base.lschnupp.value = L
    a = Series(
        DRFPMI_state('XARM'),
        OptimiseRFReadoutPhaseDC("XARM", "AS45", name='xarm'),
        DRFPMI_state('YARM'),
        OptimiseRFReadoutPhaseDC("YARM", "AS45", name='yarm'),
        DRFPMI_state('DRFPMI'),
    )
    sol = a.run(base)
    sol['yarm'].phases['AS45'], sol['xarm'].phases['AS45']

    phase = np.deg2rad(sol['yarm'].phases['AS45']-sol['xarm'].phases['AS45'])
    f = base.f2.value.eval()
    lmbda = 299792458/f
    k = 2*np.pi/lmbda
    schnupp = phase/k/2
    assert(abs(schnupp- base.lschnupp.value) < 1e-6)